from __future__ import annotations

from inspect import _ParameterKind, Parameter
import logging
import re
from typing import Any, Callable, Literal, Tuple
import sys

from openpectus.engine.commands import ContextEngineCommand, CommandArgs
from openpectus.engine.hardware import HardwareLayerBase, NullHardware, Register, RegisterDirection
from openpectus.lang.exec.base_unit import BaseUnitProvider
from openpectus.lang.exec.errors import UodValidationError
from openpectus.lang.exec.readings import (
    Reading, ReadingCollection, ReadingWithChoice, ReadingWithEntry, UodCommandDescription
)
from openpectus.lang.exec.tags import SystemTagName, Tag, TagCollection
from openpectus.lang.exec.units import get_compatible_unit_names, get_volume_units
from openpectus.lang.exec.tags_impl import AccumulatorBlockTag, AccumulatedColumnVolume, AccumulatorTag
from openpectus.protocol.models import EntryDataType, PlotConfiguration

logger = logging.getLogger(__name__)


class UnitOperationDefinitionBase:
    """ Represets the Unit Operation Definition interface used by the OpenPectus engine.
    """
    def __init__(self,
                 instrument_name: str,
                 hwl: HardwareLayerBase,
                 author_name: str,
                 author_email: str,
                 filename: str,
                 location: str,
                 tags: TagCollection,
                 readings: ReadingCollection,
                 command_factories: dict[str, UodCommandBuilder],
                 command_descriptions: dict[str, UodCommandDescription],
                 overlapping_command_names_lists: list[list[str]],
                 plot_configuration: PlotConfiguration,
                 required_roles: set[str],
                 data_log_interval_seconds: float,
                 base_unit_provider: BaseUnitProvider
                 ) -> None:
        self.instrument = instrument_name
        self.hwl = hwl
        self.author_name = author_name
        self.author_email = author_email
        self.filename = filename
        self.location = location
        self.tags = tags
        self.system_tags: TagCollection | None = None
        self.readings = readings
        self.command_factories = command_factories
        self.command_descriptions = command_descriptions
        self.command_instances: dict[str, UodCommand] = {}
        self.overlapping_command_names_lists: list[list[str]] = overlapping_command_names_lists
        self.plot_configuration = plot_configuration
        self.required_roles = required_roles
        self.data_log_interval_seconds = data_log_interval_seconds
        self.base_unit_provider: BaseUnitProvider = base_unit_provider

    def __str__(self) -> str:
        return f"UnitOperationDefinition({self.instrument=},{self.location=},hwl={str(self.hwl)})"

    @property
    def options(self) -> dict[str, str]:
        return {
            'uod_name': self.instrument,
            'uod_author_name': self.author_name,
            'uod_author_email': self.author_email,
            'uod_filename': self.filename,
            'location': self.location
        }

    def build_commands(self):
        """ Complete configuration of a validated uod. This builds the commands.
        """
        for r in self.readings:
            r.build_commands_list()

        # build command descriptions/examples documentation
        for cmd_name, cmd_builder in self.command_factories.items():
            valid_units: list[str] = []
            exclusive_options: list[str] = []
            additive_options: list[str] = []
            argument_type: Literal["float", "unknown", "none"] = "unknown"

            # Determine whether command is used for reading entry. If so, we can use the related tag's unit
            # as source of valid_units
            cmd_reading = next((r for r in self.readings if cmd_name in [c.name for c in r.commands]), None)
            if cmd_reading is not None:  # command is paired with this reading
                tag = cmd_reading.tag
                assert tag is not None
                if tag.unit is not None:
                    valid_units = get_compatible_unit_names(tag.unit)
                    if isinstance(tag.value, float):
                        argument_type = "float"

            # use the configured argument parser to determine argument and units
            regex_parser = RegexNamedArgumentParser.get_instance(cmd_builder.arg_parse_fn)
            if regex_parser is not None:
                named_groups = sorted(regex_parser.get_named_groups())
                if "number" in named_groups:  # always True for this regex
                    argument_type = "float"
                if "number_unit" in named_groups:
                    valid_units = regex_parser.get_units()

                    if cmd_reading is not None:
                        # propagate regex unit constraints back to reading's valid entry units
                        logger.debug(f"Applying regex unit constraints, {cmd_reading.valid_value_units} -> {valid_units} " +
                                     f"command {cmd_name}")
                        cmd_reading.valid_value_units = valid_units
                if "option" in named_groups:
                    exclusive_options = regex_parser.get_exclusive_options()
                    additive_options = regex_parser.get_additive_options()
            elif cmd_builder.arg_parse_fn is None:
                # no argument parser, this means no argument and units at all
                # argument_type = "none"
                valid_units = []

            desc = UodCommandDescription(name=cmd_name)
            desc.argument_type = argument_type
            desc.argument_valid_units = valid_units
            desc.argument_exclusive_options = exclusive_options
            desc.argument_additive_options = additive_options
            desc.set_docstring(cmd_builder.exec_fn.__doc__)

            self.command_descriptions[cmd_name] = desc

    def validate_configuration(self):
        """ Validates these areas:
        - Each Reading matches a defined tag
        - Each Reading Command is verified
            - Must have an exec function with an appropriate signature
            - If the command uses a regular expression as parser function, it is verified that
              the exec function arguments match the regular expression.
        - Each process value is verified
            - Checks that entry process values have matching tag value type and
              process value entry_data_type.
        """
        fatal = False

        def log_fatal(msg: str):
            nonlocal fatal
            if not fatal:
                logger.fatal("An error occured while validating the Unit Operation Definition. Pectus Engine cannot start.")
            fatal = True
            logger.fatal(msg)

        if self.instrument is None or self.instrument.strip() == "":
            log_fatal("Instrument is not configured")
        if self.hwl is None:
            log_fatal("Hardware Layer is not configured")

        self.validate_read_registers()

        if self.system_tags is None:
            log_fatal("System tags have not been set")
        else:
            tags = self.tags.merge_with(self.system_tags)
            for r in self.readings:
                try:
                    r.match_with_tags(tags)
                except UodValidationError as vex:
                    log_fatal(str(vex))

                try:
                    r.resolve_entry_type()
                except ValueError as ex:
                    log_fatal(str(ex))

            # temporary check until we can support multiple readings per tag
            tags_with_reading = []
            for r in self.readings:
                if r.tag_name in tags_with_reading:
                    log_fatal(f"Tag '{r.tag_name}' is used by multiple readings. This is not currently supported.")
                else:
                    tags_with_reading.append(r.tag_name)
        try:
            self.validate_command_signatures()
        except UodValidationError as vex:
            log_fatal("Error in command definition. " + str(vex))

        if fatal:
            sys.exit(1)

    def validate_command_signatures(self):
        """ Validate the signatures of command exec functions"""
        import inspect

        for key, builder in self.command_factories.items():
            # builder.exec_fn is the function we are validating
            if builder.exec_fn is None:
                raise UodValidationError(f"Command '{key}' must have an execution function")

            signature = inspect.signature(builder.exec_fn)

            # 'cmd' is part of the framework and is always required
            if 'cmd' not in signature.parameters.keys():
                raise UodValidationError(f"Execution function for command '{key}' is missing a 'cmd' argument")

            regex_parser = RegexNamedArgumentParser.get_instance(builder.arg_parse_fn)
            if regex_parser is not None:
                # The command is using RegexNamedArgumentParser as argument parser. In this case we can validate the exec
                # function thoroughly by checking that its argument names match the named groups in the regular expression.

                # Given a regex with named groups ['foo', 'bar'], the expected signature is:
                #   def some_name(cmd: UodCommand, foo, bar) or
                #   def some_name(cmd: UodCommand, **kvargs)
                # However, it can also be a more intricate variation where default values are used
                # for arguments that may or may not be available depending on the regex...
                # Like this:
                #   some_name(cmd: UodCommand, foo, whynot=None, nevertheless=True, bar) or
                #   some_name(cmd: UodCommand, foo, whynot=None, **myargs)
                # which are also both valid.
                try:
                    named_groups = sorted(regex_parser.get_named_groups())
                except Exception as ex:
                    raise UodValidationError(
                        f"Error in command '{key}'. Failed to find named groups in regex: '{regex_parser.regex}'. Ex: {ex}")

                # Account for each signature parameter and each named group in the regex:
                parameter_unmatched = False
                named_groups_unmatched = set(named_groups)
                # A parameter has to either:
                for name, p in signature.parameters.items():
                    # 0) be named "cmd" which is part of the framework
                    if name == "cmd":
                        continue
                    # 1) be matched by a named group in the regex
                    if name in named_groups:
                        named_groups_unmatched.discard(name)
                        continue
                    # 2) have a default value
                    if p.default is not Parameter.empty:
                        named_groups_unmatched.discard(name)
                        continue
                    # 3) be a kvargs parameter
                    if p.kind == _ParameterKind.VAR_KEYWORD:
                        named_groups_unmatched.clear()
                        continue

                    parameter_unmatched = True

                if len(named_groups_unmatched) > 0 or parameter_unmatched:
                    # Parameter not accounted for. This means we can't provide a value for it when invoking exec_fn
                    # so we fail the validation
                    raise UodValidationError(f"""Command '{key}' has an error.
The parameters for the execution function do not match those defined in the regular expression
{regex_parser.regex} .
The expected execution function signature is:
    def some_name(cmd: UodCommand, {', '.join(named_groups)})""")

            elif builder.arg_parse_fn == defaultArgumentParser:
                # Default argument parser. Require a "value" parameter or a **kvargs parameter:
                param_count = len(signature.parameters)
                last_param_name = list(signature.parameters.keys())[param_count - 1]
                last_param = signature.parameters[last_param_name]
                if param_count == 1:
                    raise UodValidationError(f"""Command '{key}' has an error.
The execution function is using the default argument parser and requires
either a 'value' argument or a '**kvargs' argument""")
                if param_count == 2:
                    if last_param.kind != _ParameterKind.VAR_KEYWORD and last_param_name != "value":
                        raise UodValidationError(f"""Command '{key}' has an error.
The execution function is using the default argument parser and requires
either a 'value' argument or a '**kvargs' argument""")

            else:
                # Custom arg parsing. we can only verify that a **kvargs parameter is present
                # if only one positional arg is present.
                param_count = len(signature.parameters)
                last_param_name = list(signature.parameters.keys())[param_count - 1]
                last_param = signature.parameters[last_param_name]
                if param_count == 2 and last_param.kind != _ParameterKind.VAR_KEYWORD:
                    raise UodValidationError(f"""Command '{key}' has an error.
The execution function is missing named arguments or a '**kvargs' argument""")

    def validate_read_registers(self):
        """ Verify that all read registers have a matching tag. """
        read_registers = [r for r in self.hwl.registers.values() if RegisterDirection.Read in r.direction]
        for r in read_registers:
            if not self.tags.has(r.name):
                raise UodValidationError(
                    f"Register '{r.name}' has read direction but no matching tag")

    def has_command_name(self, name: str) -> bool:
        """ Check whether the command name is defined in the Uod """
        if name.strip() == "":
            raise ValueError("Command name is None or empty")
        return name in self.command_factories.keys()

    def has_command_instance(self, name: str) -> bool:
        """ Check whether the Uod currently has an active instance of the named command """
        if name.strip() == "":
            raise ValueError("Command name is None or empty")
        return name in self.command_instances.keys()

    def has_any_command_instances(self) -> bool:
        return len(self.command_instances) > 0

    def create_command(self, name: str) -> UodCommand:
        """ Create a new command instance. Only one command instance with a given name can exist at a time. """
        if name.strip() == "":
            raise ValueError("Command name is None or empty")
        if self.has_command_instance(name):
            raise ValueError(f"Command {name} instance already exists")
        factory = self.command_factories.get(name)
        if factory is None:
            raise ValueError(f"Command {name} not found")
        instance = factory.build(self)
        self.command_instances[name] = instance
        return instance

    def dispose_command(self, cmd: UodCommand):
        """ Remove command from list of instances """
        if self.has_command_instance(cmd.name):
            self.command_instances.pop(cmd.name)

    def get_command(self, name: str) -> UodCommand | None:
        """ Get existing command instance by name """
        if name.strip() == "":
            raise ValueError("Command name is None or empty")
        if self.has_command_instance(name):
            return self.command_instances[name]
        raise ValueError(f"Command instance named '{name}' not found")

    def get_command_names(self) -> list[str]:
        return list(self.command_factories.keys())

    def validate_tag_name(self, tag_name: str) -> bool:
        return tag_name in self.tags.names

    def validate_command_name(self, command_name: str) -> bool:
        return command_name in self.get_command_names()

    def generate_pcode_examples(self) -> list[Tuple[str, str]]:
        """ Generate example code as tuples of (description, example_pcode). """
        examples: list[Tuple[str, str]] = []
        for description in self.command_descriptions.values():
            # Generate example from command docstring
            docstring_pcode_example = description.get_docstring_pcode()
            if not docstring_pcode_example.strip():
                logger.warning(f"{description} has no pcode example")
            else:
                examples.append((str(description), docstring_pcode_example))

            # Generate examples from arg_parse_regex if RegexCategorical
            # or RegexNumber.
            for regex_example in description.generate_pcode_examples():
                examples.append((str(description), regex_example))

        # Generate examples from readings with command_options
        for reading in self.readings:
            if reading.command_options:
                for command_option in reading.command_options.values():
                    examples.append((str(reading), command_option))

        return examples


INIT_FN = Callable[[], None]
""" Command initialization method. """

EXEC_FN = Callable[..., None]
""" Command execution function. Must take UodCommand and **kvargs as inputs. May be invoked multiple times.

NOTE: The type system does not seem to support describing the real signature (which should be something like
Callable[[UodCommand, Unpack[CommandArgs]], None]).

For this reason we validate the exec function dynamically during uod validation. This validation check must be
updated if EXEC_FN changes. It is implemented in UnitOperationDefinitionBase.verify_command_signatures().
"""

FINAL_FN = Callable[[], None]
""" Command finalization method. """

PARSE_FN = Callable[[str], CommandArgs | None]
""" Command argument parse function. Parses a string into a dictionary. Returns None on invalid input. """



class UodCommand(ContextEngineCommand[UnitOperationDefinitionBase]):
    """ Represents a command that targets hardware, such as setting a valve state.

    Uod commands are specified/implemented by using the UodCommandBuilder class.
    """
    def __init__(self, context: UnitOperationDefinitionBase, name: str) -> None:
        super().__init__(context, name)
        self.init_fn: INIT_FN | None = None
        self.exec_fn: EXEC_FN | None = None
        self.finalize_fn: FINAL_FN | None = None
        self.arg_parse_fn: PARSE_FN | None = None

    @staticmethod
    def builder() -> UodCommandBuilder:
        """ Helper method that provides a command builder object that must be used to
        specify the command. """
        return UodCommandBuilder()

    def initialize(self):
        super().initialize()

        if self.init_fn is not None:
            self.init_fn()

    def execute(self, args: CommandArgs) -> None:
        if self.exec_fn is None:
            raise ValueError(f"Command '{self.name}' has no execution function defined")

        if not isinstance(args, dict):
            raise TypeError(
                f"Invalid arguments provided to command '{self.name}'. Must be a dictionary, not {type(args).__name__}")

        super().execute(args)
        self.exec_fn(args)

    def cancel(self):
        super().cancel()

    def finalize(self):
        super().finalize()

        if self.finalize_fn is not None:
            self.finalize_fn()

        self.context.dispose_command(self)

    def parse_args(self, args: str) -> CommandArgs | None:
        """ Parse argument input to concrete values. Return None to indicate invalid arguments. """
        if self.arg_parse_fn is None:
            return {}
        else:
            return self.arg_parse_fn(args)


class UodCommandBuilder():
    """ Used to builds command specifications and as factory to instantiate commands from the specifications. """
    def __init__(self) -> None:
        self.name = ""
        self.init_fn: Callable[[UodCommand], None] | None = None
        self.exec_fn: Callable[..., None] | None = None
        self.finalize_fn: Callable[[UodCommand], None] | None = None
        self.arg_parse_fn: Callable[[str], CommandArgs | None] | None = None

    def with_name(self, name: str) -> UodCommandBuilder:
        """ Define the name of the command. Required. """
        self.name = name
        return self

    def with_init_fn(self, init_fn: Callable[[UodCommand], None]) -> UodCommandBuilder:
        """ Define an initialization function for the command. Optional. """
        self.init_fn = init_fn
        return self

    def with_exec_fn(self, exec_fn: Callable[..., None]) -> UodCommandBuilder:
        """ Define the command's execution function. Required.

        Example::

          def my_command_exec(cmd: UodCommand, **kvargs):
              # kvargs contains the result of the arg_parse function.
              pass

        """
        self.exec_fn = exec_fn
        return self

    def with_finalize_fn(self, finalize_fn: Callable[[UodCommand], None]) -> UodCommandBuilder:
        """ Define a finalizer function for the command. Optional. """
        self.finalize_fn = finalize_fn
        return self

    def with_arg_parse_fn(self, arg_parse_fn: Callable[[str], CommandArgs | None]) -> UodCommandBuilder:
        """ Define an argument parser function. Optional.

        The function is given the command argument string from the method and should parse that
        and return the result as a dictionary. The result is passed to the excution function automatically.

        If not arg_parse function is defined, no arguments are provided to the execute function.

        Example::

          def my_command_parse_args(args: str) -> dict[str, Any] | None:
            # parse the argument string and return the values as a dictionary.
            return None # if args are invalid.
        """
        self.arg_parse_fn = arg_parse_fn
        return self

    def build(self, uod: UnitOperationDefinitionBase) -> UodCommand:
        """ Construct the command """

        def arg_parse(args: str) -> CommandArgs | None:
            if self.arg_parse_fn is None:
                return {}
            return self.arg_parse_fn(args)

        def initialize() -> None:
            if self.init_fn is not None:
                return self.init_fn(c)

        def execute(args: CommandArgs) -> None:
            if self.exec_fn is not None:
                try:
                    return self.exec_fn(c, **args)
                except TypeError as te:
                    raise Exception(f"Execution function type error in uod command '{self.name}' with **args '{args}'")\
                        from te

        def finalize() -> None:
            if self.finalize_fn is not None:
                return self.finalize_fn(c)

        if self.name is None or self.name.strip() == '':
            raise ValueError("Name is not set")
        c = UodCommand(uod, self.name)
        c.init_fn = initialize
        c.exec_fn = execute
        c.finalize_fn = finalize
        c.arg_parse_fn = arg_parse
        return c


class UodBuilder():
    """ Provides a builder api to define a Unit Operation Definition """
    def __init__(self) -> None:
        self.instrument: str = ""
        self.hwl: HardwareLayerBase | None = None
        self.tags = TagCollection()
        self.command_factories: dict[str, UodCommandBuilder] = {}
        self.command_descriptions: dict[str, UodCommandDescription] = {}
        self.overlapping_command_names_lists: list[list[str]] = []
        self.readings = ReadingCollection()
        self.author_name: str = ""
        self.author_email: str = ""
        self.filename: str = ""
        self.location: str = ""
        self.plot_configuration: PlotConfiguration = PlotConfiguration.empty()
        self.required_roles: set[str] = set()
        self.data_log_interval_seconds = 5
        self.base_unit_provider: BaseUnitProvider = BaseUnitProvider()
        self.base_unit_provider.set("s", SystemTagName.BLOCK_TIME, SystemTagName.BLOCK_TIME)
        self.base_unit_provider.set("min", SystemTagName.BLOCK_TIME, SystemTagName.BLOCK_TIME)
        self.base_unit_provider.set("h", SystemTagName.BLOCK_TIME, SystemTagName.BLOCK_TIME)

    def get_logger(self):
        return logging.getLogger(f'{__name__}.user_uod')

    def validate(self):
        if len(self.instrument.strip()) == 0:
            raise ValueError("Instrument name must be set")

        if len(self.author_name.strip()) == 0 or len(self.author_email.strip()) == 0:
            raise ValueError("Author must be set")

        if len(self.filename.strip()) == 0:
            raise ValueError("Filename must be set")

        if self.hwl is None:
            raise ValueError("HardwareLayer must be set")

        return

    def with_instrument(self, instrument_name: str) -> UodBuilder:
        if len(instrument_name.strip()) == 0:
            raise ValueError("Instrument name cannot be empty")

        self.instrument = instrument_name
        return self

    def with_hardware_none(self) -> UodBuilder:
        """ Use no hardware. Mainly used for testing. """
        return self.with_hardware(NullHardware())

    def with_hardware_opcua(self, host: str) -> UodBuilder:
        """ Use OPCUA hardware """
        from openpectus.engine.opcua_hardware import OPCUA_Hardware
        hwl = OPCUA_Hardware(host)
        return self.with_hardware(hwl)

    def with_hardware_labjack(self, serial_number: str | None = None) -> UodBuilder:
        """ Use LabJack hardware """
        from openpectus.engine.labjack_hardware import Labjack_Hardware
        hwl = Labjack_Hardware(serial_number)
        return self.with_hardware(hwl)

    def with_hardware(self, hwl: HardwareLayerBase) -> UodBuilder:
        self.hwl = hwl
        return self

    def with_author(self, name: str, email: str) -> UodBuilder:
        self.author_name = name
        self.author_email = email
        return self

    def with_filename(self, filename: str) -> UodBuilder:
        self.filename = filename
        return self

    def with_location(self, location: str) -> UodBuilder:
        self.location = location
        return self

    def with_hardware_register(self, name: str, direction: RegisterDirection = RegisterDirection.Read, **options)\
            -> UodBuilder:
        """ Define a hardware register.

        Parameters:
            name (str):
                The register name. Used to match the register to a tag.
            direction:
                Specify whether the register is read, written or both
            options:
                Specify additional named options. Specific hardwares may require certain options,
                e.g. the opcua hardware requires a 'path' option.

                Two special options are callbacks named 'to_tag' and 'from_tag'. If specified, these callbacks
                convert between hardware values and tag values.
        """
        register = Register(name, direction, **options)
        if self.hwl is None:
            raise ValueError("HardwareLayber must be defined before defining a register")
        if register.name is None or register.name == "":
            raise ValueError("Register name must be non-empty")
        if register.name in self.hwl.registers.keys():
            raise ValueError(f"Register {register.name} is already defined")
        self.hwl.registers[register.name] = register
        return self

    def with_tag(self, tag: Tag) -> UodBuilder:
        # TODO Replace Tag as input with specific named variations
        # eg. with_tag_choice, with_tag_reading, with_tag
        return self._with_tag(tag)

    def _with_tag(self, tag: Tag) -> UodBuilder:
        if self.tags.has(tag.name):
            raise ValueError(f"Duplicate tag name: {tag.name}")
        self.tags.add(tag)
        return self

    def with_accumulated_volume(self, totalizer_tag_name: str) -> UodBuilder:
        if not self.tags.has(totalizer_tag_name):
            raise ValueError(f"The specified totalizer tag name '{totalizer_tag_name}' was not found")
        totalizer = self.tags[totalizer_tag_name]
        volume_units = get_volume_units()
        if totalizer.unit not in volume_units:
            raise ValueError(f"The totalizer tag '{totalizer_tag_name}' must have a volume unit")
        self._with_tag(AccumulatorTag(name=SystemTagName.ACCUMULATED_VOLUME, totalizer=totalizer))
        self._with_tag(AccumulatorBlockTag(name=SystemTagName.BLOCK_VOLUME, totalizer=totalizer))
        self.with_base_unit_provider(volume_units, SystemTagName.ACCUMULATED_VOLUME, SystemTagName.BLOCK_VOLUME)
        return self

    def with_accumulated_cv(self, cv_tag_name: str, totalizer_tag_name: str) -> UodBuilder:
        if not self.tags.has(cv_tag_name):
            raise ValueError(f"The specified column volume tag name '{cv_tag_name}' was not found")
        if not self.tags.has(totalizer_tag_name):
            raise ValueError(f"The specified totalizer tag name '{totalizer_tag_name}' was not found")
        cv = self.tags[cv_tag_name]
        totalizer = self.tags[totalizer_tag_name]
        self._with_tag(AccumulatedColumnVolume(SystemTagName.ACCUMULATED_CV, cv, totalizer))
        acc_cv = self.tags[SystemTagName.ACCUMULATED_CV]
        self._with_tag(AccumulatorBlockTag(SystemTagName.BLOCK_CV, acc_cv))
        self.with_base_unit_provider(["CV"], SystemTagName.ACCUMULATED_CV, SystemTagName.BLOCK_CV)
        return self

    def with_base_unit_provider(self, units: list[str], provider_tag_name: str, provider_block_tag_name: str):
        if provider_tag_name is None or provider_tag_name.strip() == "":
            raise ValueError("provider_tag_name is required")
        if not self.tags.has(provider_tag_name):
            raise ValueError(f"The specified provider_tag_name '{provider_tag_name}' was not found ")
        if provider_block_tag_name is None or provider_block_tag_name.strip() == "":
            raise ValueError("provider_block_tag_name is required")
        if not self.tags.has(provider_block_tag_name):
            raise ValueError(f"The specified provider_block_tag_name '{provider_block_tag_name}' was not found ")

        if len(units) == 0:
            raise ValueError("At least one unit is required")
        for unit in units:
            if self.base_unit_provider.has(unit):
                logger.warning(f"The base unit provider for unit '{unit}' was \
                               overwritten with tag name '{provider_tag_name}'")
            self.base_unit_provider.set(unit, provider_tag_name, provider_block_tag_name)
        return self

    def with_command(
            self,
            name: str,
            exec_fn: Callable[..., None],
            init_fn: Callable[[UodCommand], None] | None = None,
            finalize_fn: Callable[[UodCommand], None] | None = None,
            arg_parse_fn: Callable[[str], CommandArgs | None] | Literal["default"] | None = "default"
            ) -> UodBuilder:
        """ Define a custom unit operation command that can be executed from pcode.

        Parameters:
            name (str):
                The command name. Used in pcode to refer to the command
            exec_fn:
                The function to execute when the command is run. The function will be called
                once in each engine tick until it is completed. To signal completion, use
                `cmd.set_complete()`.
            init_fn, optional:
                An optional initialization function that is invoked once before the command runs
            finalize_fn, optional:
                An optional function that is run once when the command is completed
            arg_parse_fn, optional:
                An optional function that is used to parse the command argument in pcode to
                **kvargs that are given as named arguments to the exec function.
                A number of cases are possible:
                The default parsing function (`arg_parse_fn="default"`) passes the argument from the pcode line
                verbatim to the exec function using the argument name 'value'.
                If no parse function is set (`arg_parse_fn=None`), no arguments are passed to the exec function.
                To use a regular expression as argument parser, use the method `with_command_regex_arguments()`
                instead.
        """
        cb = UodCommandBuilder().with_name(name).with_exec_fn(exec_fn)
        if init_fn is not None:
            cb.with_init_fn(init_fn)
        if finalize_fn is not None:
            cb.with_finalize_fn(finalize_fn)
        if arg_parse_fn == "default":
            cb.with_arg_parse_fn(defaultArgumentParser)
        elif arg_parse_fn is not None:
            cb.with_arg_parse_fn(arg_parse_fn)
        if cb.name in self.command_factories.keys():
            raise ValueError(f"Duplicate command name: {cb.name}")
        self.command_factories[cb.name] = cb
        return self

    def with_command_regex_arguments(
            self,
            name: str,
            arg_parse_regex: str,
            exec_fn: Callable[..., None],
            init_fn: Callable[[UodCommand], None] | None = None,
            finalize_fn: Callable[[UodCommand], None] | None = None,
            ) -> UodBuilder:
        """ Define a custom unit operation command with a regular expression as argument parser.

        Parameters:
            name (str):
                The command name. Used in pcode to refer to the command
            arg_parse_regex (str):
                The regular expression to use for parsing. The expression must use named groups
                to specify the argument names and values.
            exec_fn:
                The function to execute when the command is run. The function will be called
                once in each engine tick until it is completed. To signal completion, use
                `cmd.set_complete()`.
            init_fn, optional:
                An optional initialization function that is invoked once before the command runs
            finalize_fn, optional:
                An optional function that is run once when the command is completed
        """
        cb = UodCommandBuilder().with_name(name).with_exec_fn(exec_fn).with_arg_parse_fn(
            RegexNamedArgumentParser(arg_parse_regex).parse)
        if init_fn is not None:
            cb.with_init_fn(init_fn)
        if finalize_fn is not None:
            cb.with_finalize_fn(finalize_fn)
        self.command_factories[cb.name] = cb
        return self

    def with_command_overlap(self, command_names: list[str]) -> UodBuilder:
        """ Specify that the given commands names overlap, i.e. at most one can run at a time. """
        if len(command_names) < 2:
            raise ValueError("To define command overlap, at least two command names are required")
        self.overlapping_command_names_lists.append(command_names)
        return self

    def with_process_value(self, tag_name: str) -> UodBuilder:
        """ Add process value that displays the given tag in the UI. """
        reading = Reading(tag_name=tag_name)
        return self._with_process_value(reading)

    def with_process_value_entry(
            self,
            tag_name: str,
            entry_data_type: EntryDataType = "auto",
            execute_command_name: str | None = None
            ) -> UodBuilder:
        """ Add process value for the given tag and enable setting its value by typing in a value in the UI.

        Parameters:
            tag_name: str
                The tag to display
            entry_data_type: "str" | "int" | "float" | "auto", optional, default="auto"
                The data type to display and validate in the UI. Default is "auto" which uses the type of the tag's
                value. The "auto" value requires that the tag has a default value with the correct type.
            execute_command_name: str: optional
                The command to execute to 'set the process value'. In principle, this can be any pcode command that takes a
                single argument as given by the user. However, the intension is that the command 'sets the process value',
                probably by writing the value to the register backing the tag.

                If not specified, it defaults to the tag name, assuming that there is a command with the same name as
                the tag that takes a single argument.
        """
        reading = ReadingWithEntry(
            tag_name=tag_name,
            entry_data_type=entry_data_type,
            execute_command_name=execute_command_name)
        return self._with_process_value(reading)

    def with_process_value_choice(self, tag_name: str, command_options: dict[str, str] | None = None) -> UodBuilder:
        """ Add process value for the given tag and enable command choices.

        Parameters:
            tag_name: str
                The tag to display. The tag's choices are also used as commands, unless
                command_options are specified.
            command_options: optional
                Dictionary that maps command names to their pcode implementation.
                If not specified, the tag's choices are used as both names and
                pcode implementation. If specified, the tag's choices are not used and the
                tag does not need to have choices defined.
        """
        if not self.tags.has(tag_name):
            raise ValueError(f"Cannot add process value choice for tag name {tag_name}. The tag name was not found")
        tag = self.tags.get(tag_name)
        if command_options is not None:
            reading = ReadingWithChoice(tag_name, command_options)
        elif tag.choices is None or len(tag.choices) == 0:
            raise ValueError(f"Cannot add process value choice for tag name {tag_name}. " +
                             "The tag has no choices defined and no options were given")
        else:
            reading = ReadingWithChoice(tag_name)
        return self._with_process_value(reading)

    def _with_process_value(self, reading: Reading) -> UodBuilder:
        self.readings.add(reading)
        return self

    def with_plot_configuration(self, plot_configuration: PlotConfiguration):
        self.plot_configuration = plot_configuration
        return self

    def with_required_roles(self, required_roles: set[str]) -> UodBuilder:
        self.required_roles = required_roles
        return self

    def with_data_log_interval_seconds(self, data_log_interval_seconds: float) -> UodBuilder:
        self.data_log_interval_seconds = data_log_interval_seconds
        return self

    def build(self) -> UnitOperationDefinitionBase:
        self.validate()

        uod = UnitOperationDefinitionBase(
            self.instrument,
            self.hwl,  # type: ignore
            self.author_name,
            self.author_email,
            self.filename,
            self.location,
            self.tags,
            self.readings,
            self.command_factories,
            self.command_descriptions,
            self.overlapping_command_names_lists,
            self.plot_configuration,
            self.required_roles,
            self.data_log_interval_seconds,
            self.base_unit_provider
        )

        return uod


def defaultArgumentParser(args: str) -> CommandArgs:
    """ The default command argument parser.

    Passes the string argument verbatim to the command exec function as a single argument
    named `value`.

    E.g. for the command pcode `MyCommand: 87`, `value` will have the value '87'.

    When using this argument parser, the exec function looks like this::

      def exec(cmd: UodCommand, value: str):
        # implementation here
        cmd.set_complete()
    """
    return {'value': args}


def unescape(re_escaped_string: str) -> str:
    """
    re.escape is used to escape options supplied to
    RegexCategorical. This function can reverses operation.
    
    assert unescape(re.escape('A B')) == 'A B'
    assert unescape(re.escape('A/B')) == 'A/B'
    """
    # Source: https://stackoverflow.com/questions/43662474/reversing-pythons-re-escape
    return re.sub(r'\\(.)', r'\1', re_escaped_string)


class RegexNamedArgumentParser():
    def __init__(self, regex: str) -> None:
        self.regex = regex

    def parse(self, args: str) -> dict[str, Any] | None:
        match = re.search(self.regex, args)
        if not match:
            return None
        return match.groupdict()

    def get_named_groups(self) -> list[str]:
        # ex: r"(?P<value>[0-9]+[.][0-9]*?|[.][0-9]+|[0-9]+) ?(?P<unit>m2)"
        # ex2: ' ?(?P<number_unit>kg)'
        # eg3: '^\\s*(?P<number>-?[0-9]+[.][0-9]*?|-?[.][0-9]+|-?[0-9]+)\\s* ?(?P<number_unit>kg|g)\\s*$'
        result = []
        p = re.compile(r"\<(?P<name>([a-zA-Z]|_)+)\>")
        for match in p.finditer(self.regex):
            result.append(match.group("name"))
        return result

    def get_units(self) -> list[str]:
        if "number_unit" not in self.get_named_groups():
            return []
        start = self.regex.index("<number_unit>") + len("<number_unit>")
        end = self.regex.index(")", start)
        result = unescape(self.regex[start: end]).split("|")
        return result

    def get_exclusive_options(self) -> list[str]:
        if "option" not in self.get_named_groups():
            return []
        start = self.regex.index("<option>") + len("<option>(")
        end = self.regex.index("|(")
        option_string = self.regex[start: end]
        result = unescape(self.regex[start: end]).split("|")
        return result

    def get_additive_options(self) -> list[str]:
        if "option" not in self.get_named_groups():
            return []
        start = self.regex.index("|(") + len("|(")
        end = self.regex.index("|\\+)+))\\s*")
        result = unescape(self.regex[start: end]).split("|")
        return result

    @staticmethod
    def get_instance(parse_func) -> RegexNamedArgumentParser | None:
        try:
            instance = getattr(parse_func, "__self__", None)
            if instance is not None and isinstance(instance, RegexNamedArgumentParser):
                return instance
        except Exception:
            return None


# Common regular expressions for use with RegexNamedArgumentParser

def RegexNumber(units: list[str] | None, non_negative: bool = False) -> str:
    """ Create a regex that parses a number with optional unit to arguments `number` and optionally `number_unit`.

    `number_unit` is only provided if one or more units are given.
    """
    sign_part = "" if non_negative else "-?"
    unit_part = " ?(?P<number_unit>" + "|".join(re.escape(unit).replace(r"/", r"\/") for unit in units) + ")" \
        if units else ""
    return rf"^\s*(?P<number>{sign_part}[0-9]+[.][0-9]*?|{sign_part}[.][0-9]+|{sign_part}[0-9]+)\s*{unit_part}\s*$"


def RegexCategorical(exclusive_options: list[str] | None = None, additive_options: list[str] | None = None) -> str:
    """ Create a regex that parses categorical text into an argument named `option`.

    Examples - exclusive::

        regex = RegexCategorical(exclusive_options=["Open", "Closed"])

        self.assertEqual(re.search(regex, "Closed").groupdict(), dict(option="Closed"))

        self.assertEqual(re.search(regex, "VA01"), None)

        self.assertEqual(re.search(regex, "Open").groupdict(), dict(option="Open"))

        self.assertEqual(re.search(regex, "Open+Closed"), None)


    Examples - exclusive and additive::

        regex = RegexCategorical(exclusive_options=['Closed'], additive_options=['VA01', 'VA02', 'VA03'])

        self.assertEqual(re.search(regex, "Closed").groupdict(), dict(option="Closed"))

        self.assertEqual(re.search(regex, "Closed+VA01"), None)

        self.assertEqual(re.search(regex, "VA01").groupdict(), dict(option="VA01"))

        self.assertEqual(re.search(regex, "VA01+VA02").groupdict(), dict(option="VA01+VA02"))

        self.assertEqual(re.search(regex, "Open+Closed"), None)
    """
    if exclusive_options is None and additive_options is None:
        raise TypeError("RegexCategorical() missing argument 'exclusive_options' or 'additive_options'.")
    exclusive_option_part = "|".join(re.escape(option) for option in exclusive_options) if exclusive_options else ""
    additive_option_part = "|".join(re.escape(option) for option in additive_options) if additive_options else ""
    return rf"^(?P<option>({exclusive_option_part}|({additive_option_part}|\+)+))\s*$"

def RegexText(allow_empty: bool = False):
    """ Parses text into an argument named `text`."""
    allow_empty_part = "*" if allow_empty else "+"
    return rf"^(?P<text>.{allow_empty_part})$"



import unittest
from uuid import UUID

from openpectus.aggregator.models import RunLog, RunLogLine
from openpectus.engine.engine import Engine
from openpectus.lang.exec.runlog import RunLogItem, RunLogItemState, RuntimeRecordStateEnum
from openpectus.test.engine.test_engine import create_engine
from openpectus.test.engine.utility_methods import (
    continue_engine, run_engine, print_runlog, print_runtime_records,
    configure_test_logger, set_engine_debug_logging
)

configure_test_logger()
set_engine_debug_logging()

class TestRunlog(unittest.TestCase):
    def setUp(self):
        self.engine: Engine = create_engine()

    def tearDown(self):
        self.engine.cleanup()

    def assert_Runlog_HasItem(self, name: str):
        runlog = self.engine.runtimeinfo.get_runlog()
        for item in runlog.items:
            if item.name == name:
                return
        item_names = [item.name for item in runlog.items]
        self.fail(f"Runlog has no item named '{name}'. It has these names:  {','.join(item_names)}")

    def assert_Runlog_HasItem_Started(self, name: str):
        runlog = self.engine.runtimeinfo.get_runlog()
        for item in runlog.items:
            if item.name == name and item.state == RunLogItemState.Started:
                return
        self.fail(f"Runlog has no item named '{name}' in Started state")

    def assert_Runlog_HasItem_Completed(self, name: str, min_times=1):
        occurrences = 0
        runlog = self.engine.runtimeinfo.get_runlog()
        for item in runlog.items:
            if item.name == name and item.state == RunLogItemState.Completed:
                occurrences += 1

        if occurrences < min_times:
            self.fail(f"Runlog item named '{name}' did not occur in Completed state at least " +
                      f"{min_times} time(s). It did occur {occurrences} time(s)")

    def assert_Runtime_HasRecord(self, name: str):
        for r in self.engine.runtimeinfo.records:
            if r.name == name:
                return
        self.fail(f"Runtime has no record named '{name}'")

    def assert_Runtime_HasRecord_Started(self, name: str):
        for r in self.engine.runtimeinfo.records:
            if r.name == name and r.has_state(RuntimeRecordStateEnum.Started):
                return
        self.fail(f"Runtime has no record named '{name}' in state Started")

    def assert_Runtime_HasRecord_Completed(self, name: str):
        for r in self.engine.runtimeinfo.records:
            if r.name == name and r.has_state(RuntimeRecordStateEnum.Started):
                return
        self.fail(f"Runtime has no record named '{name}' in state Completed")

    def test_start_complete_UodCommand(self):
        e = self.engine

        run_engine(e, "Reset", 3)
        self.assert_Runlog_HasItem_Started("Reset")

        continue_engine(e, 5)

        print_runtime_records(e)
        print_runlog(e)

        self.assert_Runlog_HasItem_Completed("Reset")

        self.assert_Runtime_HasRecord_Started("Reset")
        self.assert_Runtime_HasRecord_Completed("Reset")

    def test_start_complete_InstructionUodCommand(self):
        e = self.engine

        run_engine(e, "Mark: A", 3)

        self.assert_Runtime_HasRecord_Started("Mark: A")
        self.assert_Runtime_HasRecord_Completed("Mark: A")

        self.assert_Runlog_HasItem_Completed("Mark: A")

    def test_start_complete_EngineInternalCommand(self):
        e = self.engine

        cmd = "Increment run counter"
        run_engine(e, cmd, 5)

        print_runtime_records(e)

        self.assert_Runtime_HasRecord_Started(cmd)
        self.assert_Runtime_HasRecord_Completed(cmd)

        self.assert_Runlog_HasItem_Completed(cmd)

    def test_Watch(self):
        e = self.engine

        cmd = """
Mark: a
Watch: Run Counter > -1
    Mark: b
Mark: c"""
        run_engine(e, cmd, 5)

        cmd = "Mark: b"
        self.assert_Runtime_HasRecord_Started(cmd)
        self.assert_Runtime_HasRecord_Completed(cmd)
        self.assert_Runlog_HasItem_Completed(cmd)

    def test_Alarm_single_invocation(self):
        e = self.engine

        cmd = """
Alarm: Run Counter < 5
    Mark: b
    Increment run counter
"""
        run_engine(e, cmd, 5)

        cmd = "Mark: b"
        self.assert_Runtime_HasRecord_Started(cmd)
        self.assert_Runtime_HasRecord_Completed(cmd)
        self.assert_Runlog_HasItem_Completed(cmd, 1)

    def test_Alarm_multiple_invocations(self):
        e = self.engine

        cmd = """
Alarm: Run Counter < 3
    Mark: b
    Increment run counter
"""
        run_engine(e, cmd, 10)

        # print_runtime_records(e)
        # print_runlog(e)

        cmd = "Mark: b"

        self.assert_Runtime_HasRecord_Started(cmd)
        self.assert_Runtime_HasRecord_Completed(cmd)
        self.assert_Runlog_HasItem_Completed(cmd, 3)

        # print_runtime_records(e)

    def test_runlog_cancel_watch(self):
        e = self.engine
        program = """
Watch: Block Time > .3s
    Mark: Foo
"""
        run_engine(e, program, 3)
        item_name = "Watch: Block Time > .3s"
        self.assert_Runlog_HasItem(item_name)

        runlog = e.runtimeinfo.get_runlog()
        item = next((i for i in runlog.items if i.name == item_name), None)
        assert isinstance(item, RunLogItem)
        exec_id = item.id

        # verify that its runlog item is cancelable
        self.assertEqual(item.cancellable, True)
        self.assertEqual(item.cancelled, False)
        print_runtime_records(e)
        print_runlog(e, "pre-cancel")

        # cancel it. interpreter needs a tick to process it
        e.cancel_instruction(UUID(exec_id))
        continue_engine(e, 5)

        # fetch the updated (rebuilt from runtime records) runlog
        runlog = e.runtimeinfo.get_runlog()
        item = next((i for i in runlog.items if i.name == item_name), None)
        assert isinstance(item, RunLogItem)

        # verify it was cancelled
        print_runtime_records(e, "post-cancel")
        print_runlog(e, "post-cancel")

        self.assertEqual(item.cancelled, True)
        self.assertEqual(item.cancellable, False)
        self.assertEqual([], e.interpreter.get_marks())

        # needs end to be set or the frontend won't know the state is and end state
        self.assertIsNotNone(item.end)

    def test_runlog_force_watch(self):
        e = self.engine
        program = """
Watch: Block Time > .3s
    Mark: Foo
"""
        run_engine(e, program, 3)
        item_name = "Watch: Block Time > .3s"
        self.assert_Runlog_HasItem(item_name)

        runlog = e.runtimeinfo.get_runlog()
        item = next((i for i in runlog.items if i.name == item_name), None)
        assert isinstance(item, RunLogItem)
        exec_id = item.id

        # verify that its runlog item is forcible
        self.assertEqual(item.forcible, True)
        self.assertEqual(item.forced, False)
        print_runtime_records(e)
        print_runlog(e, "pre-force")

        # force it. interpreter needs a tick to process it
        e.force_instruction(UUID(exec_id))
        continue_engine(e, 1)

        # fetch the updated (rebuilt from runtime records) runlog
        runlog = e.runtimeinfo.get_runlog()
        item = next((i for i in runlog.items if i.name == item_name), None)
        assert isinstance(item, RunLogItem)

        # verify it was forced
        print_runtime_records(e, "post-force")
        print_runlog(e, "post-force")

        self.assertEqual(item.forced, True)
        self.assertEqual(item.forcible, False)
        self.assertEqual(['Foo'], e.interpreter.get_marks())

    def test_iteration_modification(self):
        # test iteration_modification as used in engine._execute_uod_command
        # doesn't seem to be a problem. little weird that the condition is not
        # reevaluated during iteration.
        foos = ['foo', 'bar', 'baz']
        bazs = []
        result = []
        for x in [f for f in foos if f not in bazs]:
            if x == "bar":
                bazs.append('baz')
            result.append(x)
        self.assertEqual(['foo', 'bar', 'baz'], result)

    def test_runlog_cancel_uod_command(self):
        e = self.engine
        # start uod command
        run_engine(e, "Reset", 3)
        self.assert_Runlog_HasItem_Started("Reset")

        runlog = e.runtimeinfo.get_runlog()
        self.assertEqual(1, len([i for i in runlog.items if i.name == "Reset"]))
        item = next((i for i in runlog.items if i.name == "Reset"), None)
        assert isinstance(item, RunLogItem)
        exec_id = item.id

        # verify that its runlog item is cancelable
        print_runlog(e)
        self.assertEqual(item.cancellable, True)
        self.assertEqual(item.cancelled, False)

        print_runtime_records(e)
        assert item.progress is not None
        self.assertEqual(item.progress > 0.0, True)

        # cancel it. engine needs a tick to process it
        e.cancel_instruction(UUID(exec_id))
        continue_engine(e, 1)

        # fetch the updated (rebuilt from runtime records) runlog
        runlog = e.runtimeinfo.get_runlog()
        item = next((i for i in runlog.items if i.name == "Reset"), None)
        assert isinstance(item, RunLogItem)

        # verify it was cancelled
        print_runlog(e)
        print_runtime_records(e, "post-cancel")
        self.assertEqual(item.cancelled, True)
        self.assertEqual(item.cancellable, False)

        # assert we only have one item
        self.assertEqual(1, len([i for i in runlog.items if i.name == "Reset"]))

    def test_runlog_force_alarm(self):
        e = self.engine
        program = """
Alarm: Block Time > .3s
    Mark: Foo
"""
        run_engine(e, program, 3)
        item_name = "Alarm: Block Time > .3s"
        self.assert_Runlog_HasItem(item_name)

        runlog = e.runtimeinfo.get_runlog()
        item = next((i for i in runlog.items if i.name == item_name), None)
        assert isinstance(item, RunLogItem)
        exec_id = item.id

        # verify that its runlog item is forcible
        self.assertEqual(item.forcible, True)
        self.assertEqual(item.forced, False)
        print_runtime_records(e)
        print_runlog(e, "pre-force")

        # force it. interpreter needs a tick to process it
        e.force_instruction(UUID(exec_id))
        continue_engine(e, 1)

        # fetch the updated (rebuilt from runtime records) runlog
        runlog = e.runtimeinfo.get_runlog()
        item = next((i for i in runlog.items if i.name == item_name), None)
        assert isinstance(item, RunLogItem)

        # verify it was forced
        print_runtime_records(e, "post-force")
        print_runlog(e, "post-force")

        self.assertEqual(item.forced, True)
        self.assertEqual(item.forcible, False)
        self.assertEqual(['Foo'], e.interpreter.get_marks())

    def test_runlog_watch_in_alarm_body_runs_in_each_alarm_instance(self):
        e = self.engine
        program = """
Alarm: Block Time > 0s
    Watch: Block Time > 0.5s
        Mark: A
    Wait: 0.5s
"""
        run_engine(e, program, 4)
        alarm_item_name = "Alarm: Block Time > 0s"
        watch_item_name = "Watch: Block Time > 0.5s"

        print_runtime_records(e, "start")

        self.assert_Runlog_HasItem(alarm_item_name)
        self.assert_Runlog_HasItem(watch_item_name)

        continue_engine(e, 6)
        self.assertEqual(['A'], e.interpreter.get_marks())

        continue_engine(e, 10)  # not sure how long to wait - but surely this is enough
        self.assert_Runlog_HasItem_Completed(alarm_item_name, 2)  # verify we waited long enough
        self.assertEqual(['A', 'A'], e.interpreter.get_marks())

    def test_runlog_force_Mark_w_threshold_is_not_forcible(self):
        e = self.engine
        program = """
2 Mark: A
"""
        run_engine(e, program, 4)

        mark_name = "Mark: A"
        self.assert_Runlog_HasItem(mark_name)
        runlog = e.runtimeinfo.get_runlog()
        item = next(item for item in runlog.items if item.name == mark_name)
        assert item is not None

        self.assertEqual(item.forcible, False)
        self.assertEqual(item.forced, False)

    def test_runlog_force_Mark_without_threshold_is_not_forcible(self):
        e = self.engine
        program = """
Mark: A
"""
        run_engine(e, program, 4)

        mark_name = "Mark: A"
        self.assert_Runlog_HasItem(mark_name)
        runlog = e.runtimeinfo.get_runlog()
        item = next(item for item in runlog.items if item.name == mark_name)
        assert item is not None

        self.assertEqual(item.forcible, False)
        self.assertEqual(item.forced, False)

    def test_wait_progress_EngineInternalCommand(self):
        e = self.engine

        cmd = "Wait: 0.5s"
        item_name = cmd

        run_engine(e, cmd, 5)
        print_runtime_records(e)

        self.assert_Runtime_HasRecord_Started(cmd)
        self.assert_Runlog_HasItem(item_name)

        runlog = e.runtimeinfo.get_runlog()
        item = next((i for i in runlog.items if i.name == item_name), None)
        assert item is not None and item.progress is not None
        # after 5-2 = 3 ticks we're 0.6 percent done
        self.assertAlmostEqual(item.progress, 0.6, delta=0.05)
        self.assertEqual(item.forcible, True)
        self.assertEqual(item.cancellable, False)

        continue_engine(e, 1)
        runlog = e.runtimeinfo.get_runlog()
        item = next((i for i in runlog.items if i.name == item_name), None)
        assert item is not None and item.progress is not None
        # after 1 more tick we're 0.8 percent done
        self.assertAlmostEqual(item.progress, 0.8, delta=0.05)

        continue_engine(e, 2)
        self.assert_Runlog_HasItem_Completed(cmd)


    def test_runlog_Wait_is_forcible(self):
        e = self.engine
        program = """
Wait: 0.5s
"""
        run_engine(e, program, 4)

        cmd_name = "Wait: 0.5s"
        self.assert_Runlog_HasItem(cmd_name)
        runlog = e.runtimeinfo.get_runlog()
        item = next(item for item in runlog.items if item.name == cmd_name)
        assert item is not None
        exec_id = UUID(item.id)

        self.assertEqual(item.forcible, True)
        self.assertEqual(item.forced, False)

        e.force_instruction(exec_id)
        continue_engine(e, 1)

        # get updated runlog
        runlog = e.runtimeinfo.get_runlog()
        item = next(item for item in runlog.items if item.name == cmd_name)
        assert item is not None

        self.assertEqual(item.forcible, False)
        self.assertEqual(item.forced, True)

    def test_runlog_aggregator_eq(self):
        l1 = RunLogLine(id="id", command_name="c", start=1, end=None, progress=None,
                        start_values=[], end_values=[],
                        forcible=False, forced=False, cancellable=False, cancelled=False)

        l2 = RunLogLine(id="id", command_name="c", start=1, end=None, progress=None,
                        start_values=[], end_values=[],
                        forcible=False, forced=False, cancellable=False, cancelled=False)

        l3 = RunLogLine(id="id2", command_name="c", start=1, end=None, progress=None,
                        start_values=[], end_values=[],
                        forcible=False, forced=False, cancellable=False, cancelled=False)

        l4 = RunLogLine(id="id", command_name="c", start=1, end=None, progress=0.2,
                        start_values=[], end_values=[],
                        forcible=False, forced=False, cancellable=False, cancelled=False)

        self.assertTrue(l1 == l2)
        self.assertTrue(l1 != l3)
        self.assertTrue(l1 != l4)

        rl1 = RunLog(lines=[l1])
        rl2 = RunLog(lines=[l2])
        rl3 = RunLog(lines=[l3])
        rl4 = RunLog(lines=[l4])

        self.assertTrue(rl1 == rl2)
        self.assertTrue(rl1 != rl3)
        self.assertTrue(rl1 != rl4)


if __name__ == "__main__":
    unittest.main()



import unittest

from openpectus.lang.grammar.pgrammar import PGrammar


def get_token_texts(code):
    p = PGrammar()
    p.parse(code)
    tokens = list(p.tokens)
    return list(t.text for t in tokens)


class LexerTest(unittest.TestCase):
    def assertPTokenCountGE(self, code: str, expected_min_count: int):

        def print(s):
            pass
        print(f"\nCode: '{code}'")
        print("Tokens:")
        p = PGrammar()
        p.parse(code)
        tokens = p.tokens
        for t in tokens:
            print(t)
        self.assertGreaterEqual(len(tokens), expected_min_count)

    def test_block(self):
        code = "block:foo"
        self.assertPTokenCountGE(code, 1)

        code = "block: foo "
        self.assertPTokenCountGE(code, 1)

        code = "\nblock:foo#cmt\n"
        self.assertPTokenCountGE(code, 1)

    def test_command(self):
        code = "foo #no comment"
        self.assertPTokenCountGE(code, 1)

    def test_condition(self):
        code = "foo > 3"
        self.assertEqual(['foo', ' ', '>', ' ', '3'], get_token_texts(code))

    def test_condition_2(self):
        code = "foo=3"
        self.assertEqual(['foo', '=', '3'], get_token_texts(code))

    def test_condition_negative_value(self):
        code = "foo=-3"
        self.assertEqual(['foo', '=', '-', '3'], get_token_texts(code))

    def test_condition_unit(self):
        code = "foo > 3 ml"
        self.assertEqual(['foo', ' ', '>', ' ', '3', ' ', 'ml'], get_token_texts(code))

    def test_watch_condition(self):
        code = "watch:foo > 3 ml"
        self.assertEqual(['watch', ':', 'foo', ' ', '>', ' ', '3', ' ', 'ml'], get_token_texts(code))

    def test_pause(self):
        code = "Pause"
        self.assertEqual(['Pause'], get_token_texts(code))

    def test_pause_w_arg(self):
        code = "Pause: 2 s"
        self.assertEqual(['Pause', ':', ' ', '2', ' ', 's'], get_token_texts(code))

    def test_hold(self):
        code = "Hold"
        self.assertEqual(['Hold'], get_token_texts(code))

    def test_hold_w_arg(self):
        code = "Hold: 2 s"
        self.assertEqual(['Hold', ':', ' ', '2', ' ', 's'], get_token_texts(code))

    def test_hold_w_arg2(self):
        code = "Hold: 2.0 min"
        self.assertEqual(['Hold', ':', ' ', '2.0', ' ', 'min'], get_token_texts(code))

    def test_mark_name_percent(self):
        code = "Mark: 2%"
        self.assertEqual(['Mark', ':', ' ', '2', '%'], get_token_texts(code))

        code = "Mark: B %"
        self.assertEqual(['Mark', ':', ' ', 'B', ' ', '%'], get_token_texts(code))


if __name__ == "__main__":
    unittest.main()



from typing import List

from antlr4 import InputStream, CommonTokenStream, Token, ParserRuleContext
from antlr4.tree.Tree import ParseTree, ParseTreeWalker

from openpectus.lang.exec.analyzer import EnrichAnalyzer
from openpectus.lang.grammar.codegen.pcodeLexer import pcodeLexer
from openpectus.lang.grammar.codegen.pcodeParser import pcodeParser
from openpectus.lang.model.pprogram import PProgram
from openpectus.lang.grammar.pprogrambuilder import PProgramBuilder


class PGrammar:
    def __init__(self) -> None:
        self.lexer = pcodeLexer
        self.parser = pcodeParser

    def parse(self, code: str):  # code: TextIOWrapper):
        input = InputStream(code)
        self.lexer = pcodeLexer(input)
        stream = CommonTokenStream(self.lexer)
        self.parser = pcodeParser(stream)

    def build_model(self, skip_enrich_analyzers=False) -> PProgram:
        self.tree: pcodeParser.ProgramContext = self.parser.program()  # type: ignore
        listener = PProgramBuilder()
        walker = ParseTreeWalker()
        walker.walk(listener, self.tree)
        program = listener.get_program()
        if not skip_enrich_analyzers:
            EnrichAnalyzer().analyze(program)
        program.collect_errors()
        return program

    @property
    def tokens(self) -> List[Token]:
        return list(self.lexer.getAllTokens())  # type: ignore

    @staticmethod
    def _recursive(aRoot: ParseTree, buf: List[str], offset: int, ruleNames: List[str]):
        for _ in range(offset):
            buf.append("  ")

        buf.append(aRoot.getText() + " | " + type(aRoot).__name__ + "\n")  # type: ignore

        if isinstance(aRoot, ParserRuleContext):
            if aRoot.children is not None:
                for child in aRoot.children:
                    PGrammar._recursive(child, buf, offset + 1, ruleNames)

    def printSyntaxTree(self, root: ParseTree, initial_newline: bool = True):
        buf: List[str] = []
        self._recursive(root, buf, 0, list(self.parser.ruleNames))
        if initial_newline:
            print()
        print("".join(buf))



import logging
import os
import contextlib

import uvicorn
from openpectus.aggregator.routers.auth import UserRolesDependency
from fastapi import FastAPI, APIRouter
from fastapi.routing import APIRoute
from openpectus import __version__
from openpectus.aggregator.aggregator_message_handlers import AggregatorMessageHandlers
from openpectus.aggregator.data import database
from openpectus.aggregator.deps import _create_aggregator
from openpectus.aggregator.frontend_publisher import FrontendPublisher
from openpectus.aggregator.routers import process_unit, recent_runs, auth, version
from openpectus.aggregator.spa import SinglePageApplication
from openpectus.protocol.aggregator_dispatcher import AggregatorDispatcher


class AggregatorServer:
    default_title = "Open Pectus Aggregator"
    default_frontend_dist_dir = os.path.join(os.path.dirname(__file__), "frontend-dist")
    default_host = "127.0.0.1"
    default_port = 9800
    default_db_filename = "open_pectus_aggregator.sqlite3"
    default_db_path = os.path.join(os.getcwd(), default_db_filename)
    default_secret = ""

    def __init__(self, title: str = default_title, host: str = default_host, port: int = default_port,
                 frontend_dist_dir: str = default_frontend_dist_dir, db_path: str = default_db_path,
                 secret: str = default_secret):
        self.title = title
        self.host = host
        self.port = port
        self.frontend_dist_dir = frontend_dist_dir
        self.db_path = db_path
        if not os.path.exists(frontend_dist_dir):
            raise FileNotFoundError("{frontend_dist_dir} not found.")
        self.dispatcher = AggregatorDispatcher()
        self.publisher = FrontendPublisher()
        self.aggregator = _create_aggregator(self.dispatcher, self.publisher, secret)
        _ = AggregatorMessageHandlers(self.aggregator)
        self.setup_fastapi([self.dispatcher.router, self.publisher.router, version.router])
        self.init_db()

    def setup_fastapi(self, additional_routers: list[APIRouter] = []):
        api_prefix = "/api"

        def custom_generate_unique_id(route: APIRoute):
            return f"{route.name}"

        self.fastapi = FastAPI(title=self.title,
                               version=__version__,
                               contact=dict(name="Open Pectus",
                                            url="https://github.com/Open-Pectus/Open-Pectus"),
                               generate_unique_id_function=custom_generate_unique_id,
                               lifespan=self.lifespan)
        self.fastapi.include_router(process_unit.router, prefix=api_prefix, dependencies=[UserRolesDependency])
        self.fastapi.include_router(recent_runs.router, prefix=api_prefix, dependencies=[UserRolesDependency])
        self.fastapi.include_router(auth.router, prefix="/auth")
        for route in additional_routers:
            self.fastapi.include_router(route)
        self.fastapi.mount("/", SinglePageApplication(directory=self.frontend_dist_dir))

    def init_db(self):
        database.configure_db(f"sqlite:///{self.db_path}")
        self.fastapi.add_middleware(database.DBSessionMiddleware)

    def start(self):
        uvicorn.run(self.fastapi, host=self.host, port=self.port, log_level=logging.WARNING)

    @contextlib.asynccontextmanager
    async def lifespan(self, app):
        yield
        await self.dispatcher.shutdown()



import logging
from datetime import UTC, datetime, timedelta, timezone
from socket import gethostname
from typing import Iterable, Sequence

from openpectus import __version__
from openpectus.aggregator.data.models import (
    RecentEngine,
    ProcessValueType,
    RecentRunErrorLog, RecentRunMethodAndState, RecentRun, PlotLogEntryValue,
    RecentRunPlotConfiguration, RecentRunRunLog,
    get_ProcessValueType_from_value,
    PlotLog, PlotLogEntry
)
from openpectus.aggregator.models import TagValue, ReadingInfo, EngineData
from sqlalchemy import select
from sqlalchemy.orm import Session

from openpectus.protocol.models import SystemTagName

logger = logging.getLogger(__name__)


class RepositoryBase():
    def __init__(self, db_session: Session) -> None:
        self.db_session = db_session


class PlotLogRepository(RepositoryBase):
    def create_plot_log(self, engine_data: EngineData, run_id: str):
        plot_log = PlotLog()
        plot_log.engine_id = engine_data.engine_id
        plot_log.run_id = run_id

        def map_reading(reading: ReadingInfo):
            entry = PlotLogEntry()
            entry.plot_log = plot_log
            entry.value_type = ProcessValueType.NONE
            entry.name = reading.tag_name
            return entry

        entries = map(map_reading, engine_data.readings)

        self.db_session.add(plot_log)
        self.db_session.add_all(entries)
        self.db_session.commit()

        # store info from existing tags
        for tag in engine_data.tags_info.map.values():
            self.store_new_tag_info(engine_data.engine_id, run_id, tag)

    def store_new_tag_info(self, engine_id: str, run_id: str, tag: TagValue):
        existing_plot_log_entry = self.get_plot_log_entry(engine_id, run_id, tag)
        if existing_plot_log_entry is None:
            logger.debug(f'tag {tag.name} was not found in readings')
            return
        existing_plot_log_entry.value_unit = tag.value_unit
        existing_plot_log_entry.value_type = get_ProcessValueType_from_value(tag.value)
        self.db_session.add(existing_plot_log_entry)
        self.db_session.commit()

    def get_plot_log(self, run_id: str) -> PlotLog | None:
        return self.db_session.scalar(
            select(PlotLog)
            .join(PlotLog.entries)
            .join(PlotLogEntry.values)
            .where(PlotLog.run_id == run_id)
        )

    def get_plot_log_entry(self, engine_id: str, run_id: str, tag: TagValue) -> PlotLogEntry | None:
        return self.db_session.scalar(
            select(PlotLogEntry)
            .join(PlotLog)
            .where(PlotLog.engine_id == engine_id)
            .where(PlotLogEntry.name == tag.name)
            .where(PlotLog.run_id == run_id)
        )

    def get_plot_log_entries(self, engine_id: str, run_id: str) -> Iterable[PlotLogEntry]:
        return self.db_session.scalars(
            select(PlotLogEntry)
            .join(PlotLog)
            .where(PlotLog.engine_id == engine_id)
            .where(PlotLog.run_id == run_id)
        ).all()

    def store_tag_values(self, engine_id: str, run_id: str, tags: list[TagValue]):
        plot_log_entries = self.get_plot_log_entries(engine_id, run_id)

        def map_tag_to_entry_value(tag: TagValue):
            plot_log_entry = next((entry for entry in plot_log_entries if entry.name == tag.name), None)
            if plot_log_entry is None:
                return None
            return PlotLogEntryValue(
                plot_log_entry_id=plot_log_entry.id,
                tick_time=tag.tick_time,
                value_str=tag.value if isinstance(tag.value, str) else None,
                value_float=tag.value if isinstance(tag.value, float) else None,
                value_int=tag.value if isinstance(tag.value, int) else None,
            )

        db_models = (entry_value for entry_value in map(map_tag_to_entry_value, tags) if entry_value is not None)
        self.db_session.add_all(db_models)
        self.db_session.commit()


class RecentRunRepository(RepositoryBase):
    def store_recent_run(self, engine_data: EngineData):
        """ Store a recent run. Requires that engine_data contain run_data. """
        if not engine_data.has_run():
            raise ValueError('missing run_data when trying to store recent run')
        assert engine_data.has_run(), "Run data must be set when saving recent run"
        run_id = engine_data.run_data.run_id
        recent_run = RecentRun()
        recent_run.engine_id = engine_data.engine_id
        recent_run.run_id = run_id
        recent_run.engine_computer_name = engine_data.computer_name
        recent_run.engine_version = engine_data.engine_version
        recent_run.engine_hardware_str = engine_data.hardware_str
        recent_run.aggregator_version = __version__
        recent_run.aggregator_computer_name = gethostname()
        recent_run.uod_name = engine_data.uod_name
        recent_run.uod_author_name = engine_data.uod_author_name
        recent_run.uod_author_email = engine_data.uod_author_email
        recent_run.uod_filename = engine_data.uod_filename
        recent_run.started_date = engine_data.run_data.run_started
        recent_run.completed_date = datetime.now(timezone.utc)
        current_contributors = set(recent_run.contributors or [])
        for user_name in engine_data.contributors:  # append contributers from this 'session'
            current_contributors.add(user_name)
        recent_run.contributors = list(current_contributors)  # must assign new instance in json field
        recent_run.required_roles = list(engine_data.required_roles)

        method_and_state = RecentRunMethodAndState()
        method_and_state.run_id = run_id
        method_and_state.method = engine_data.method
        method_and_state.state = engine_data.method_state

        plot_configuration = RecentRunPlotConfiguration()
        plot_configuration.run_id = run_id
        plot_configuration.plot_configuration = engine_data.plot_configuration

        run_log = RecentRunRunLog()
        run_log.run_id = run_id
        run_log.run_log = engine_data.run_data.runlog

        error_log = RecentRunErrorLog()
        error_log.run_id = run_id
        error_log.error_log = engine_data.error_log

        self.db_session.add(recent_run)
        self.db_session.add(method_and_state)
        self.db_session.add(plot_configuration)
        self.db_session.add(run_log)
        self.db_session.add(error_log)
        self.db_session.commit()

    def get_by_run_id(self, run_id: str) -> RecentRun | None:
        return self.db_session.scalar(select(RecentRun).where(RecentRun.run_id == run_id))

    def get_by_engine_id(self, engine_id: str) -> Iterable[RecentRun]:
        return self.db_session.scalars(
            select(RecentRun)
            .where(RecentRun.engine_id == engine_id)
        ).all()

    def get_all(self) -> Iterable[RecentRun]:
        return self.db_session.scalars(select(RecentRun)).all()

    def get_method_and_state_by_run_id(self, run_id: str):
        return self.db_session.scalar(select(RecentRunMethodAndState).where(RecentRunMethodAndState.run_id == run_id))

    def get_plot_configuration_by_run_id(self, run_id: str):
        return self.db_session.scalar(
            select(RecentRunPlotConfiguration.plot_configuration).where(RecentRunPlotConfiguration.run_id == run_id)
        )

    def get_run_log_by_run_id(self, run_id: str):
        return self.db_session.scalar(select(RecentRunRunLog.run_log).where(RecentRunRunLog.run_id == run_id))

    def get_error_log_by_run_id(self, run_id: str):
        return self.db_session.scalar(select(RecentRunErrorLog.error_log).where(RecentRunErrorLog.run_id == run_id))


class RecentEngineRepository(RepositoryBase):
    def get_recent_engines(self) -> Sequence[RecentEngine]:
        last_month = datetime.now() - timedelta(days=30)
        return self.db_session.scalars(
            select(RecentEngine)
            .where(RecentEngine.last_update > last_month)
        ).all()

    def get_recent_engine_by_engine_id(self, engine_id: str) -> RecentEngine | None:
        return self.db_session.scalar(select(RecentEngine).where(RecentEngine.engine_id == engine_id))

    def store_recent_engine(self, engine_data: EngineData):
        if engine_data.engine_id == "":
            raise ValueError("Missing/empty engine_id when trying to store recent_engine")
        existing = self.get_recent_engine_by_engine_id(engine_data.engine_id)
        if existing is None:
            recent_engine = RecentEngine()
        else:
            recent_engine = existing
        recent_engine.engine_id = engine_data.engine_id
        if engine_data.has_run():
            assert engine_data.run_data is not None
            recent_engine.run_id = engine_data.run_data.run_id
            recent_engine.run_started = engine_data.run_data.run_started
            contributors = set(recent_engine.contributors)
            for c in engine_data.contributors:
                contributors.add(c)
            recent_engine.contributors = list(contributors)
        else:
            recent_engine.run_id = None
            recent_engine.run_started = None
        recent_engine.run_stopped = None
        recent_engine.name = f"{engine_data.computer_name} ({engine_data.uod_name})"
        recent_engine.location = engine_data.location
        recent_engine.last_update = datetime.now(UTC)
        recent_engine.required_roles = engine_data.required_roles
        system_tag = engine_data.tags_info.get(SystemTagName.SYSTEM_STATE)
        recent_engine.system_state = str(system_tag.value) if system_tag is not None else ""
        if system_tag is None:
            logger.warning("The SYSTEM_STATE tag value was not available when saving recent_engine")
        self.db_session.add(recent_engine)
        self.db_session.commit()



from contextlib import contextmanager
import logging
import threading
import time
import unittest
from typing import Any, Generator
from openpectus.lang.exec.errors import UodValidationError
from openpectus.lang.exec.tags_impl import ReadingTag, SelectTag

import openpectus.protocol.models as Mdl
import pint
from openpectus.engine.engine import Engine
from openpectus.engine.hardware import HardwareLayerBase, Register, RegisterDirection
from openpectus.engine.models import EngineCommandEnum, MethodStatusEnum, SystemStateEnum, SystemTagName
from openpectus.lang.exec.runlog import RuntimeRecordStateEnum
from openpectus.lang.exec.tags import Tag, TagDirection
from openpectus.lang.exec.uod import (
    UnitOperationDefinitionBase,
    UodCommand,
    UodBuilder,
    RegexNamedArgumentParser,
    RegexNumber,
)
from openpectus.test.engine.utility_methods import (
    continue_engine, run_engine,
    configure_test_logger, set_engine_debug_logging, set_interpreter_debug_logging,
    print_runlog, print_runtime_records
)
from typing_extensions import override


configure_test_logger()
set_engine_debug_logging()


logging.getLogger("openpectus.lang.exec.pinterpreter").setLevel(logging.INFO)

# pint takes forever to initialize - long enough
# to throw off timing of the first instruction.
# so we initialize it first
_ = pint.Quantity("0 sec")


def get_queue_items(q) -> list[Tag]:
    items = []
    while not q.empty():
        items.append(q.get())
    return items


def create_test_uod() -> UnitOperationDefinitionBase:
    def reset(cmd: UodCommand, **kvargs) -> None:
        count = cmd.get_iteration_count()
        max_ticks = 5
        if count == 0:
            cmd.context.tags.get("Reset").set_value("Reset", time.time())
        elif count > max_ticks:
            cmd.context.tags.get("Reset").set_value("N/A", time.time())
            cmd.set_complete()
        else:
            progress = count/max_ticks
            cmd.set_progress(progress)

    def cmd_with_args(cmd: UodCommand, **kvargs) -> None:
        print("arguments: " + str(kvargs))
        cmd.set_complete()

    def cmd_arg_parse(args: str) -> dict[str, Any] | None:
        if args is None or args == "" or args == "FAIL":
            return None
        result = {}
        for i, item in enumerate(args.split(",")):
            result[f"item{i}"] = item.strip()
        return result

    def cmd_regex(cmd: UodCommand, number, number_unit) -> None:
        cmd.context.tags["CmdWithRegex_Area"].set_value(number + number_unit, time.monotonic())
        cmd.set_complete()

    def overlap_exec(cmd: UodCommand, **kvargs) -> None:
        count = cmd.get_iteration_count()
        if count >= 9:
            cmd.set_complete()

    builder = (
        UodBuilder()
        .with_instrument("TestUod")
        .with_author("Test Author", "test@openpectus.org")
        .with_filename(__file__)
        .with_hardware(TestHW())
        .with_location("Test location")
        .with_hardware_register("FT01", RegisterDirection.Both, path="Objects;2:System;2:FT01")
        .with_hardware_register(
            "Reset",
            RegisterDirection.Both,
            path="Objects;2:System;2:RESET",
            from_tag=lambda x: 1 if x == "Reset" else 0,
            to_tag=lambda x: "Reset" if x == 1 else "N/A",
        )
        # Readings
        .with_tag(ReadingTag("FT01", "L/h"))
        .with_tag(SelectTag("Reset", value="N/A", unit=None, choices=["Reset", "N/A"]))
        .with_tag(Tag("Danger", value=True, unit=None, direction=TagDirection.Output, safe_value=False))
        .with_command(name="Reset", exec_fn=reset)
        .with_command(name="CmdWithArgs", exec_fn=cmd_with_args, arg_parse_fn=cmd_arg_parse)
        .with_command(name="overlap1", exec_fn=overlap_exec)
        .with_command(name="overlap2", exec_fn=overlap_exec)
        .with_command_overlap(['overlap1', 'overlap2'])
        .with_tag(Tag("CmdWithRegex_Area", value=""))
        .with_command_regex_arguments(
            name="CmdWithRegexArgs",
            arg_parse_regex=RegexNumber(units=['m2']),
            exec_fn=cmd_regex)
    )
    uod = builder.build()
    uod.hwl.connect()
    return uod


def create_engine(uod: UnitOperationDefinitionBase | None = None) -> Engine:
    if uod is None:
        uod = create_test_uod()
    e = Engine(uod)
    return e


@contextmanager
def create_engine_context(uod: UnitOperationDefinitionBase) -> Generator[Engine, Any, None]:
    e = create_engine(uod)
    try:
        yield e
    finally:
        e.cleanup()


class TestEngineSetup(unittest.TestCase):

    def test_create_engine(self):
        e = create_engine()
        self.assertIsNotNone(e)
        e.cleanup()

    def test_configure_uod(self):
        uod = create_test_uod()
        e = Engine(uod)

        self.assertTrue(len(uod.command_factories) > 0)
        self.assertTrue(len(uod.instrument) > 0)
        self.assertTrue(len(uod.tags) > 0)

        e.cleanup()


class TestEngine(unittest.TestCase):

    def setUp(self):
        self.engine: Engine = create_engine()

    def tearDown(self):
        self.engine.cleanup()

    @unittest.skip("Potentially affects other tests.")
    def test_engine_start(self):
        e = self.engine

        t = threading.Thread(target=e.run)
        t.daemon = True
        t.start()

        time.sleep(1)
        e._running = False
        t.join()

    @unittest.skip("Potentially affects other tests.")
    def test_engine_started_runs_scan_cycle(self):
        e = self.engine

        t = threading.Thread(target=e._run, daemon=True)  # dont configure twice
        t.start()

        # assert loop running
        time.sleep(0.5)
        self.assertTrue(e._running)

        time.sleep(0.5)
        e._running = False
        t.join()

    def test_read_process_image_marks_assigned_tags_dirty(self):
        e = self.engine

        hwl: TestHW = e.uod.hwl  # type: ignore
        self.assertIsInstance(hwl, TestHW)
        # set non-default values to trigger notification
        hwl.register_values["FT01"] = 1
        hwl.register_values["Reset"] = 1

        run_engine(e, "", 1)
        time.sleep(0.1)

        # assert tags marked dirty
        dirty_names = [t.name for t in get_queue_items(e.tag_updates)]
        self.assertTrue("FT01" in dirty_names)
        self.assertTrue("Reset" in dirty_names)

    def test_execute_command_marks_assigned_tags_dirty(self):
        e = self.engine
        reset_tag = e.uod.tags["Reset"]

        self.assertEqual("N/A", reset_tag.get_value())

        run_engine(e, "Reset", 3)
        self.assertEqual("Reset", reset_tag.get_value())

        # assert tags marked dirty
        dirty_names = [t.name for t in get_queue_items(e.tag_updates)]
        self.assertTrue("Reset" in dirty_names)

    def test_read_process_image_sets_assigned_tag_values(self):
        e = self.engine

        # set hw values
        hwl: TestHW = e.uod.hwl  # type: ignore
        self.assertIsInstance(hwl, TestHW)
        hwl.register_values["FT01"] = 87
        hwl.register_values["Reset"] = 0

        # assert values read
        e.read_process_image()
        # numerical value is set
        self.assertEqual(87, e.uod.tags["FT01"].get_value())
        # mapped value is set
        self.assertEqual("N/A", e.uod.tags["Reset"].get_value())

    def test_notify_tag_updates_publishes_dirty_tag(self):
        e = self.engine
        t = e.uod.tags["FT01"]
        e._uod_listener.notify_change(t.name)

        e.notify_tag_updates()

        self.assertEqual(1, e.tag_updates.qsize())
        t_updated = e.tag_updates.get()
        self.assertEqual(t_updated.name, t.name)

    def test_write_process_values_writes_data_to_registers(self):
        e = self.engine

        run_engine(e, "", 1)
        time.sleep(0.1)

        # set hw values
        hwl: TestHW = e.uod.hwl  # type: ignore
        self.assertIsInstance(hwl, TestHW)
        hwl.register_values["FT01"] = 87

        # assert values read
        e.read_process_image()
        self.assertEqual(87, e.uod.tags["FT01"].get_value())

        # modify tag values
        e.uod.tags["FT01"].set_value(22, e._tick_time)
        e.uod.tags["Reset"].set_value("Reset", e._tick_time)

        e.write_process_image()

        # assert values written to registers
        self.assertEqual(22, hwl.register_values["FT01"])
        self.assertEqual(1, hwl.register_values["Reset"])

    def test_engine_start_writes_safe_values_to_hw(self):
        e = self.engine

        # get Dange tag which has an initial value of True
        danger_tag = e.tags["Danger"]
        self.assertEqual(True, danger_tag.value)

        # trigger startup to write safe values
        e._run()

        # safe value has now been set
        self.assertEqual(False, danger_tag.value)

    def test_uod_command_w_arguments(self):
        e = self.engine
        program = "CmdWithArgs: a, b,c ,  d"
        run_engine(e, program, 10)

        self.assertEqual(e.tags[SystemTagName.METHOD_STATUS].get_value(), MethodStatusEnum.OK)

    def test_uod_command_w_arguments_fail(self):
        e = self.engine
        program = "CmdWithArgs: FAIL"
        with self.assertRaises(ValueError):
            run_engine(e, program, 10)
        self.assertEqual(e.tags[SystemTagName.METHOD_STATUS].get_value(), MethodStatusEnum.ERROR)

    def test_uod_command_w_regex_arguments(self):
        e = self.engine
        program = "CmdWithRegexArgs: 3.14 m2"
        run_engine(e, program, 10)

        self.assertEqual("3.14m2", e.tags["CmdWithRegex_Area"].get_value())

    def test_RegexNamedArgumentParser_parse(self):
        regex = r"(?P<value>[0-9]+[.][0-9]*?|[.][0-9]+|[0-9]+) ?(?P<unit>m2)"
        parser = RegexNamedArgumentParser(regex)
        result = parser.parse("3.14m2")

        assert result is not None
        self.assertEqual(2, len(result))
        self.assertEqual("3.14", result["value"])
        self.assertEqual("m2", result["unit"])

    def test_RegexNamedArgumentParser_get_named_groups(self):
        regex = r"(?P<value>[0-9]+[.][0-9]*?|[.][0-9]+|[0-9]+) ?(?P<unit>m2)"
        parser = RegexNamedArgumentParser(regex)
        self.assertEqual(['value', 'unit'], parser.get_named_groups())

    def test_uod_verify_command_signatures_regex(self):

        def no_args(cmd: UodCommand):
            pass

        def kvargs(cmd: UodCommand, **kvargs):
            pass

        def two_args(cmd: UodCommand, value, unit):
            pass

        def two_args_bad_names(cmd: UodCommand, one, two):
            pass

        def two_args_def(cmd: UodCommand, value, unit=None):
            pass

        exec_functions = {
            'no_args': no_args,
            'kvargs': kvargs,
            'two_args': two_args,
            'two_args_bad_names': two_args_bad_names,
            'two_args_def': two_args_def,
        }
        regexes = {
            'two_parms': "(?P<value>[0-9]+[.][0-9]*?|[.][0-9]+|[0-9]+) ?(?P<unit>m2)",
            'one_parms': "(?P<value>[0-9]+[.][0-9]*?|[.][0-9]+|[0-9]+) ?",
            'no_parms': "",
        }

        def test(exec_name, regex_name, expect_success: bool = True):
            with self.subTest(exec_name + ", " + regex_name + ", " + ("success" if expect_success else "fail")):
                exec_fn = exec_functions[exec_name]
                regex = regexes[regex_name]
                uod = (
                    UodBuilder()
                    .with_instrument("foo")
                    .with_author("Test Author", "test@openpectus.org")
                    .with_filename(__file__)
                    .with_hardware_none()
                    .with_command_regex_arguments("bar", regex, exec_fn)
                    .build()
                )

                if expect_success:
                    uod.validate_command_signatures()
                else:
                    self.assertRaises(UodValidationError, uod.validate_command_signatures)

        test('kvargs', "two_parms")
        test('no_args', "no_parms")
        test('two_args', "two_parms")
        test('two_args_def', "one_parms")

        test('no_args', "two_parms", False)
        test('two_args_bad_names', "two_parms", False)
        test('two_args', "one_parms", False)

    def test_uod_verify_command_signatures_default(self):

        def no_args(cmd: UodCommand):
            pass

        def kvargs(cmd: UodCommand, **kvargs):
            pass

        def value_arg(cmd: UodCommand, value):
            pass

        def value_arg_default(cmd: UodCommand, value=None):
            pass

        exec_functions = {
            'no_args': no_args,
            'kvargs': kvargs,
            'value_arg': value_arg,
            'value_arg_default': value_arg_default,
        }

        def test(exec_name, expect_success: bool = True):
            with self.subTest(exec_name + ", " + ("success" if expect_success else "fail")):
                uod = (
                    UodBuilder()
                    .with_instrument("foo")
                    .with_author("Test Author", "test@openpectus.org")
                    .with_filename(__file__)
                    .with_hardware_none()
                    .with_command("bar", exec_fn=exec_functions[exec_name])
                    .build()
                )

                if expect_success:
                    uod.validate_command_signatures()
                else:
                    self.assertRaises(UodValidationError, uod.validate_command_signatures)

        test('kvargs')
        test('value_arg')
        test("value_arg_default")

        test('no_args', False)

    def test_uod_invoke_command_signatures(self):

        def no_args(cmd: UodCommand):
            pass

        def kvargs(cmd: UodCommand, **kvargs):
            pass

        def value_arg(cmd: UodCommand, value):
            pass

        def value_arg_default(cmd: UodCommand, value=None):
            pass

        exec_functions = {
            'no_args': no_args,
            'kvargs': kvargs,
            'value_arg': value_arg,
            'value_arg_default': value_arg_default,
        }

        def test(exec_name, parsed_args: dict[str, Any], expect_success: bool = True):
            with self.subTest(exec_name + ", " + str(parsed_args) + ", " + ("success" if expect_success else "fail")):
                exec_fn = exec_functions[exec_name]
                try:
                    exec_fn(cmd=None, **parsed_args)
                    if not expect_success:
                        raise AssertionError("Expected execution error but none occurred")
                except TypeError:
                    if expect_success:
                        raise AssertionError("Expected no execution error but one occurred")
                except Exception as ex:
                    raise AssertionError("Did not expect exception") from ex

        test('no_args', {})
        test('kvargs', {'bar': 1})
        test('kvargs', {})
        test('value_arg', {'value': 23})
        test("value_arg_default", {})
        test("value_arg_default", {'value': 34})

        test('no_args', {'bar': 1}, False)
        test('value_arg', {}, False)
        test('value_arg', {'foo': 34}, False)
        test('value_arg', {'value': 1, 'foo': 34}, False)
        test("value_arg_default", {'foo': 34}, False)

    def test_uod_command_can_execute_valid_command(self):
        e = self.engine
        run_engine(e, "", 3)
        self.assertEqual("N/A", e.uod.tags["Reset"].get_value())

        e.inject_code("Reset")
        continue_engine(e, 1)

        self.assertEqual("Reset", e.uod.tags["Reset"].get_value())

    def test_uod_commands_multiple_iterations(self):
        e = self.engine

        self.assertEqual("N/A", e.uod.tags["Reset"].get_value())

        run_engine(e, "Reset", 2)
        self.assertEqual("Reset", e.uod.tags["Reset"].get_value())

        # Reset takes 5 ticks to revert
        continue_engine(e, 5)
        self.assertEqual("Reset", e.uod.tags["Reset"].get_value())

        continue_engine(e, 1)
        self.assertEqual("N/A", e.uod.tags["Reset"].get_value())

        continue_engine(e, 25)
        self.assertEqual("N/A", e.uod.tags["Reset"].get_value())

        print_runtime_records(e)
        print_runlog(e)

    def test_concurrent_uod_commands(self):
        e = self.engine
        run_engine(e, """
Reset
Reset
""", 5)

        records = e.interpreter.runtimeinfo.records

        print_runtime_records(e)

        r = records[2]
        self.assertEqual("Reset", r.name)
        self.assertTrue(r.has_state(RuntimeRecordStateEnum.Started))
        self.assertTrue(r.has_state(RuntimeRecordStateEnum.Cancelled))
        self.assertTrue(not r.has_state(RuntimeRecordStateEnum.Completed))

        r = records[3]
        self.assertEqual("Reset", r.name)
        self.assertTrue(r.has_state(RuntimeRecordStateEnum.Started))
        self.assertTrue(not r.has_state(RuntimeRecordStateEnum.Cancelled))
        self.assertTrue(not r.has_state(RuntimeRecordStateEnum.Completed))

        continue_engine(e, 5)
        print_runtime_records(e)
        self.assertTrue(r.has_state(RuntimeRecordStateEnum.Completed))

    def test_overlapping_uod_commands(self):
        e = self.engine
        run_engine(e, """
overlap1
overlap2
""", 5)

        rs = e.runtimeinfo.records

        print_runtime_records(e)

        r = rs[2]
        self.assertEqual("overlap1", r.name)
        self.assertTrue(r.has_state(RuntimeRecordStateEnum.Started))
        self.assertTrue(r.has_state(RuntimeRecordStateEnum.Cancelled))
        self.assertTrue(not r.has_state(RuntimeRecordStateEnum.Forced))
        self.assertTrue(not r.has_state(RuntimeRecordStateEnum.Completed))

        continue_engine(e, 1)
        print_runtime_records(e)

        r = rs[3]
        self.assertEqual("overlap2", r.name)
        self.assertTrue(r.has_state(RuntimeRecordStateEnum.Started))
        self.assertTrue(not r.has_state(RuntimeRecordStateEnum.Cancelled))
        self.assertTrue(not r.has_state(RuntimeRecordStateEnum.Forced))
        self.assertTrue(not r.has_state(RuntimeRecordStateEnum.Completed))

        continue_engine(e, 1)
        print_runtime_records(e)

        self.assertTrue(not r.has_state(RuntimeRecordStateEnum.Completed))

        print_runlog(e)

    def test_runlog_block(self):
        program = """
Block: A
    End block
Mark: X
"""
        e = self.engine
        run_engine(e, program, 3)

        rs = e.runtimeinfo.records

        print_runtime_records(e)

        self.assertEqual(3, len(rs))
        self.assertEqual("Block: A", rs[2].name)
        self.assertFalse(rs[0].has_state(RuntimeRecordStateEnum.Started))

        continue_engine(e, 1)
        self.assertTrue(rs[2].has_state(RuntimeRecordStateEnum.Started))

        continue_engine(e, 1)
        self.assertTrue(rs[2].has_state(RuntimeRecordStateEnum.Completed))

        print_runtime_records(e)

    def test_runlog_watch(self):
        program = """
Watch: Block Time > 0.2s
    Mark: A
Mark: X
"""
        e = self.engine
        run_engine(e, program, 4)

        print_runtime_records(e)

        r = e.interpreter.runtimeinfo.records[2]
        self.assertEqual("Watch: Block Time > 0.2s", r.name)
        self.assertTrue(r.has_state(RuntimeRecordStateEnum.AwaitingInterrrupt))
        self.assertTrue(r.has_state(RuntimeRecordStateEnum.AwaitingCondition))
        # self.assertFalse(r.has_state(RuntimeRecordStateEnum.Started))

        continue_engine(e, 1)
        self.assertTrue(r.has_state(RuntimeRecordStateEnum.Started))

        continue_engine(e, 1)
        self.assertTrue(r.has_state(RuntimeRecordStateEnum.Completed))

        print_runtime_records(e)

    def test_runlog_uod_commands(self):
        e = self.engine
        run_engine(e, """
Reset
Reset
Reset
""", 10)

        print_runtime_records(e)

    @unittest.skip("Not implemented yet")
    def test_runlog_watch_forced(self):
        # TODO how do we represent the instruction we want to force?
        # user picks a runlog line to force
        # - how do we get to cmd or node from that?
        # - line number? or we assign uuids to entries?
        # - how about a modified program? should use version numbers
        #   to handle program modifications and runlog implications

        raise NotImplementedError()

    def test_internal_command_can_execute_valid_command(self):
        e = self.engine

        self.assertEqual(0, e._system_tags["Run Counter"].get_value())

        e.schedule_execution("Start")
        e.tick(0, 0)

        e.inject_code("Increment run counter")
        e.tick(0, 0)

        self.assertEqual(1, e._system_tags["Run Counter"].get_value())

        print_runlog(e)

    def test_clocks(self):
        e = self.engine

        self.assertTrue(e._system_tags.has("Clock"))

        # raise NotImplementedError()

    def test_set_and_run_program(self):
        program = """
Mark: A
Mark: B
Mark: C
"""
        e = self.engine
        run_engine(e, program, 10)

        self.assertEqual(['A', 'B', 'C'], e.interpreter.get_marks())

    def test_set_and_continue_program(self):
        program = """
Mark: A
"""
        e = self.engine
        run_engine(e, program, 10)
        self.assertEqual(['A', ], e.interpreter.get_marks())

        program = """
Mark: B
Mark: C
"""
        e.set_method(Mdl.Method.from_pcode(pcode=program))
        continue_engine(e, 10)

        self.assertEqual(['B', 'C'], e.interpreter.get_marks())

    @unittest.skip("not implemented, needs input")
    def test_get_status(self):
        # program progress
        # status that enable ProcessUnit:
        # name, state=ProcessUnitState (READY, IN_PROGRESS , NOT_ONLINE) , location="H21.5", runtime_msec=189309,
        raise NotImplementedError()

    def test_get_runlog(self):
        e = self.engine
        run_engine(e, "Reset", 3)

        print_runtime_records(e)

        items = e.get_runlog().items
        self.assertEqual(1, len(items))
        self.assertEqual("Reset", items[0].name)

        # assert has tags
        for item in items:
            start_values = item.start_values
            assert start_values is not None
            self.assertTrue(start_values.has("Reset"))

    def test_clocks_start_stop(self):
        e = self.engine

        clock = e._system_tags.get(SystemTagName.CLOCK)
        self.assertEqual(0.0, clock.as_number())

        e._runstate_started = True
        e.tick(1, 1)

        clock_value = clock.as_number()
        self.assertGreater(clock_value, 0.0)

        e._runstate_started = False
        e.tick(2, 1)

        self.assertEqual(clock_value, clock.as_number())


    # --- RunState ---


    def test_runstate_start(self):
        e = self.engine

        e.schedule_execution("Start")
        e.tick(1, 1)

        self.assertTrue(e._runstate_started)

    def test_runstate_stop(self):
        set_engine_debug_logging()

        e = self.engine
        e.schedule_execution("Start")

        e.tick(1, 1)
        self.assertTrue(e._runstate_started)
        system_state_tag = e._system_tags[SystemTagName.SYSTEM_STATE]
        self.assertEqual(SystemStateEnum.Running, system_state_tag.get_value())

        e.schedule_execution("Stop")
        e.tick(1, 1)
        e.tick(1, 1)

        self.assertFalse(e._runstate_started)
        self.assertEqual(SystemStateEnum.Stopped, system_state_tag.get_value())

    def test_runstate_pause(self):
        e = self.engine

        def tick():
            tick_time = time.time()
            e.tick(tick_time, 0.1)
            time.sleep(0.1)

        e.schedule_execution("Start")
        tick()

        self.assertTrue(e._runstate_started)
        system_state_tag = e._system_tags[SystemTagName.SYSTEM_STATE]
        process_time_tag = e._system_tags[SystemTagName.PROCESS_TIME]
        self.assertEqual(SystemStateEnum.Running, system_state_tag.get_value())

        danger_tag = e.uod.tags["Danger"]
        self.assertTrue(danger_tag.get_value())

        e.schedule_execution("Pause")
        tick()

        # now paused so we can get the paused value
        pause_process_time = process_time_tag.as_number()

        # run a few ticks while paused
        tick()
        tick()

        self.assertTrue(e._runstate_started)
        self.assertTrue(e._runstate_paused)
        self.assertEqual(SystemStateEnum.Paused, system_state_tag.get_value())

        # process time is stopped
        self.assertEqual(pause_process_time, process_time_tag.as_number())

        # Pause has triggered safe-mode
        self.assertFalse(danger_tag.get_value())

    def test_runstate_unpause(self):
        e = self.engine

        def tick():
            tick_time = time.time()
            e.tick(tick_time, 0.1)
            time.sleep(0.1)

        # apply start and pause
        e.schedule_execution("Start")
        tick()

        self.assertTrue(e._runstate_started)
        system_state_tag = e._system_tags[SystemTagName.SYSTEM_STATE]
        process_time_tag = e._system_tags[SystemTagName.PROCESS_TIME]
        self.assertEqual(SystemStateEnum.Running, system_state_tag.get_value())

        danger_tag = e.uod.tags["Danger"]
        self.assertTrue(danger_tag.get_value())

        e.schedule_execution("Pause")
        tick()

        pause_process_time = process_time_tag.as_number()

        e.schedule_execution("Unpause")
        tick()

        self.assertEqual(e._runstate_started, True)
        self.assertEqual(e._runstate_paused, False)
        self.assertEqual(system_state_tag.get_value(), SystemStateEnum.Running)

        # Unpause applies pre-pause values
        self.assertEqual(danger_tag.get_value(), True)

        # needs an additional tick in system state Running before process time resumes
        tick()

        # process time is resumed
        process_time_increment = process_time_tag.as_number() - pause_process_time
        self.assertTrue(process_time_increment > 0.0)

    def test_runstate_hold(self):
        def tick():
            tick_time = time.time()
            e.tick(tick_time, 0.1)
            time.sleep(0.1)

        e = self.engine
        e.schedule_execution("Start")

        tick()
        self.assertTrue(e._runstate_started)
        system_state_tag = e._system_tags[SystemTagName.SYSTEM_STATE]
        self.assertEqual(SystemStateEnum.Running, system_state_tag.get_value())

        danger_tag = e.uod.tags["Danger"]
        self.assertTrue(danger_tag.get_value())

        e.schedule_execution("Hold")
        tick()

        self.assertTrue(e._runstate_started)
        self.assertTrue(e._runstate_holding)
        self.assertEqual(SystemStateEnum.Holding, system_state_tag.get_value())

        # Hold does not trigger safe-mode so danger is still True
        self.assertTrue(danger_tag.get_value())

    @unittest.skip("not implemented")
    def test_runstate_unhold(self):
        raise NotImplementedError()


    def test_pause_w_duration(self):
        e = self.engine
        program = """
Mark: a
Pause: .5s
Mark: b
"""
        run_engine(e, program, 5)
        self.assertTrue(e._runstate_started)
        self.assertTrue(e._runstate_paused)
        system_state_tag = e._system_tags[SystemTagName.SYSTEM_STATE]

        self.assertEqual(SystemStateEnum.Paused, system_state_tag.get_value())

        continue_engine(e, 5)

        self.assertEqual(SystemStateEnum.Running, system_state_tag.get_value())

    def test_hold_w_duration(self):
        e = self.engine
        program = """
Mark: a
Hold: .5s
Mark: b
"""
        run_engine(e, program, 7)
        self.assertTrue(e._runstate_started)
        self.assertTrue(e._runstate_holding)
        system_state_tag = e._system_tags[SystemTagName.SYSTEM_STATE]

        self.assertEqual(SystemStateEnum.Holding, system_state_tag.get_value())

        continue_engine(e, 5)

        self.assertEqual(SystemStateEnum.Running, system_state_tag.get_value())

    def test_wait(self):
        e = self.engine
        program = """
Mark: a
Wait: .5s
Mark: b
"""
        run_engine(e, program, 2)
        self.assertEqual([], e.interpreter.get_marks())

        continue_engine(e, 1)
        self.assertEqual(['a'], e.interpreter.get_marks())

        continue_engine(e, 5)
        self.assertEqual(['a'], e.interpreter.get_marks())

        continue_engine(e, 3)
        self.assertEqual(['a', 'b'], e.interpreter.get_marks())

    # --- Safe values ---


    def test_safe_values_apply(self):

        def tick():
            tick_time = time.time()
            e.tick(tick_time, 0.1)
            time.sleep(0.1)

        e = self.engine
        e.schedule_execution("Start")

        tick()

        danger_tag = e.uod.tags["Danger"]
        self.assertTrue(danger_tag.get_value())

        e._apply_safe_state()

        self.assertFalse(danger_tag.get_value())

    def test_restart_can_stop(self):
        set_engine_debug_logging()
        set_interpreter_debug_logging()
        program = """
Mark: A
Restart
"""
        e = self.engine
        run_engine(e, program, 4)

        # when no commands need to be stopped, restart immediately moves to Stopped
        system_state = e.tags[SystemTagName.SYSTEM_STATE]
        self.assertEqual(system_state.get_value(), SystemStateEnum.Restarting)


    # --- Totalizers ---


    def test_totalizer_base_units_no_accumulator(self):
        e = self.engine

        with self.subTest("allows_time_unit"):
            run_engine(e, "Base: s\n0.1 Mark: A", 5)
            self.assertEqual(e.tags[SystemTagName.METHOD_STATUS].get_value(), MethodStatusEnum.OK)

        with self.subTest("disallows_volume_unit"):
            with self.assertRaises(ValueError):
                run_engine(e, "Base: L\n0.1 Mark: A", 5)
            self.assertEqual(e.tags[SystemTagName.METHOD_STATUS].get_value(), MethodStatusEnum.ERROR)

    def test_totalizer_base_units_with_accumulator_volume(self):
        self.engine.cleanup()  # dispose the test default engine

        uod = (UodBuilder()
               .with_instrument("TestUod")
               .with_author("Test Author", "test@openpectus.org")
               .with_filename(__file__)
               .with_hardware(TestHW())
               .with_location("Test location")
               .with_tag(ReadingTag("Meter", "L"))
               .with_accumulated_volume(totalizer_tag_name="Meter")
               .build())

        with self.subTest("allows_time_unit"):
            with create_engine_context(uod) as e:
                run_engine(e, "Base: s\n0.1 Mark: A", 5)
                self.assertEqual(e.tags[SystemTagName.METHOD_STATUS].get_value(), MethodStatusEnum.OK)

        with self.subTest("allows_volume_unit"):
            with create_engine_context(uod) as e:
                run_engine(e, "Base: L\n0.1 Mark: A", 5)
                self.assertEqual(e.tags[SystemTagName.METHOD_STATUS].get_value(), MethodStatusEnum.OK)

        with self.subTest("disallows_cv_unit"):
            with create_engine_context(uod) as e:
                with self.assertRaises(ValueError):
                    run_engine(e, "Base: CV\n0.1 Mark: A", 5)
                self.assertEqual(e.tags[SystemTagName.METHOD_STATUS].get_value(), MethodStatusEnum.ERROR)

    def test_totalizer_base_units_with_accumulator_cv(self):
        self.engine.cleanup()  # dispose the test default engine

        uod = (UodBuilder()
               .with_instrument("TestUod")
               .with_author("Test Author", "test@openpectus.org")
               .with_filename(__file__)
               .with_hardware(TestHW())
               .with_location("Test location")
               .with_tag(ReadingTag("Meter", "L"))
               .with_tag(ReadingTag("CV", "L"))
               .with_accumulated_cv(cv_tag_name="CV", totalizer_tag_name="Meter")
               .build())

        uod.tags["CV"].set_value(1.0, 0)

        with self.subTest("allows_time_unit"):
            with create_engine_context(uod) as e:
                run_engine(e, "Base: s\n0.1 Mark: A", 5)
                self.assertEqual(e.tags[SystemTagName.METHOD_STATUS].get_value(), MethodStatusEnum.OK)

        with self.subTest("disallows_volume_unit"):
            with create_engine_context(uod) as e:
                with self.assertRaises(ValueError):
                    run_engine(e, "Base: L\n0.1 Mark: A", 5)
                self.assertEqual(e.tags[SystemTagName.METHOD_STATUS].get_value(), MethodStatusEnum.ERROR)

        with self.subTest("allows_cv_unit"):
            with create_engine_context(uod) as e:
                run_engine(e, "Base: CV\n0.1 Mark: A", 5)
                self.assertEqual(e.tags[SystemTagName.METHOD_STATUS].get_value(), MethodStatusEnum.OK)


    def test_accumulated_volume(self):
        self.engine.cleanup()  # dispose the test default engine

        uod = (UodBuilder()
               .with_instrument("TestUod")
               .with_author("Test Author", "test@openpectus.org")
               .with_filename(__file__)
               .with_hardware(TestHW())
               .with_location("Test location")
               .with_tag(CalculatedLinearTag("calc", "L"))
               .with_accumulated_volume(totalizer_tag_name="calc")
               .build())

        program = """
Base: s
1 Mark: A
        """
        with create_engine_context(uod) as e:
            acc_vol = e.tags[SystemTagName.ACCUMULATED_VOLUME]
            run_engine(e, program, 1)

            self.assertEqual(acc_vol.as_float(), 0.0)
            self.assertEqual(acc_vol.unit, "L")

            t0 = e._clock.get_time()
            continue_engine(e, 10)
            t1 = e._clock.get_time()

            self.assertAlmostEqual(t1 - t0, 1, delta=0.1)
            self.assertAlmostEqual(acc_vol.as_float(), 1, delta=0.1)

    def test_accumulated_block_volume(self):
        self.engine.cleanup()  # dispose the test default engine

        set_interpreter_debug_logging()

        uod = (UodBuilder()
               .with_instrument("TestUod")
               .with_author("Test Author", "test@openpectus.org")
               .with_filename(__file__)
               .with_hardware(TestHW())
               .with_location("Test location")
               .with_tag(CalculatedLinearTag("calc", "L"))
               .with_accumulated_volume(totalizer_tag_name="calc")
               .build())

        program = """
Base: s
Block: A
    0.45 End block
0.45 Mark: A
        """
        with create_engine_context(uod) as e:
            acc_vol = e.tags[SystemTagName.ACCUMULATED_VOLUME]
            block_vol = e.tags[SystemTagName.BLOCK_VOLUME]
            block = e.tags[SystemTagName.BLOCK]

            run_engine(e, program, 1)
            self.assertEqual(block.get_value(), None)

            self.assertEqual(acc_vol.as_float(), 0.0)
            self.assertEqual(acc_vol.unit, "L")
            self.assertEqual(block_vol.as_float(), 0.0)
            self.assertEqual(block_vol.unit, "L")

            continue_engine(e, 2)  # Blank + Base
            self.assertEqual(block.get_value(), None)
            self.assertAlmostEqual(acc_vol.as_float(), 0.2, delta=0.1)
            self.assertAlmostEqual(block_vol.as_float(), 0.2, delta=0.1)

            continue_engine(e, 1)  # Block
            self.assertEqual(block.get_value(), "A")
            self.assertAlmostEqual(acc_vol.as_float(), 0.3, delta=0.1)
            self.assertAlmostEqual(block_vol.as_float(), 0.1, delta=0.1)

            continue_engine(e, 5)
            self.assertEqual(block.get_value(), "A")
            self.assertAlmostEqual(acc_vol.as_float(), 0.8, delta=0.1)
            self.assertAlmostEqual(block_vol.as_float(), 0.6, delta=0.1)

            continue_engine(e, 1)
            self.assertEqual(block.get_value(), None)
            # acc_vol keeps counting
            self.assertAlmostEqual(acc_vol.as_float(), 0.9, delta=0.1)
            # block_vol is reset to value before block A - so it matches acc_vol again
            self.assertAlmostEqual(block_vol.as_float(), 0.9, delta=0.1)

    def test_accumulated_column_volume(self):
        self.engine.cleanup()  # dispose the test default engine

        uod = (UodBuilder()
               .with_instrument("TestUod")
               .with_author("Test Author", "test@openpectus.org")
               .with_filename(__file__)
               .with_hardware(TestHW())
               .with_location("Test location")
               .with_tag(ReadingTag("CV", "L"))
               .with_tag(CalculatedLinearTag("calc", "L"))
               .with_accumulated_cv(cv_tag_name="CV", totalizer_tag_name="calc")
               .build())

        program = """
Base: s
1 Mark: A
"""
        with create_engine_context(uod) as e:
            cv = e.tags["CV"]
            cv.set_value(2.0, 0)
            acc_cv = e.tags[SystemTagName.ACCUMULATED_CV]
            run_engine(e, program, 1)

            self.assertEqual(acc_cv.as_float(), 0.0)
            self.assertEqual(acc_cv.unit, "CV")

            t0 = e._clock.get_time()
            continue_engine(e, 10)
            t1 = e._clock.get_time()

            self.assertAlmostEqual(t1 - t0, 1, delta=0.1)
            self.assertAlmostEqual(acc_cv.as_float(), 1/2, delta=0.1)

    def test_accumulated_column_volume_base(self):
        self.engine.cleanup()  # dispose the test default engine

        uod = (UodBuilder()
               .with_instrument("TestUod")
               .with_author("Test Author", "test@openpectus.org")
               .with_filename(__file__)
               .with_hardware(TestHW())
               .with_location("Test location")
               .with_tag(ReadingTag("CV", "L"))
               .with_tag(CalculatedLinearTag("calc", "L"))
               .with_accumulated_cv(cv_tag_name="CV", totalizer_tag_name="calc")
               .build())

        program = """
Base: CV
0.5 Mark: A
"""
        with create_engine_context(uod) as e:
            cv = e.tags["CV"]
            cv.set_value(2.0, 0)
            acc_cv = e.tags[SystemTagName.ACCUMULATED_CV]
            run_engine(e, program, 1)

            self.assertEqual(acc_cv.as_float(), 0.0)
            self.assertEqual(acc_cv.unit, "CV")

            continue_engine(e, 4)
            self.assertAlmostEqual(acc_cv.as_float(), 0.2, delta=0.1)
            self.assertEqual([], e.interpreter.get_marks())

            continue_engine(e, 4)
            self.assertAlmostEqual(acc_cv.as_float(), 0.4, delta=0.1)
            self.assertEqual([], e.interpreter.get_marks())

            continue_engine(e, 4)
            self.assertAlmostEqual(acc_cv.as_float(), 0.6, delta=0.1)
            self.assertEqual(['A'], e.interpreter.get_marks())

    def test_accumulated_column_volume_watch(self):
        self.engine.cleanup()  # dispose the test default engine

        uod = (UodBuilder()
               .with_instrument("TestUod")
               .with_author("Test Author", "test@openpectus.org")
               .with_filename(__file__)
               .with_hardware(TestHW())
               .with_location("Test location")
               .with_tag(ReadingTag("CV", "L"))
               .with_tag(CalculatedLinearTag("calc", "L"))
               .with_accumulated_cv(cv_tag_name="CV", totalizer_tag_name="calc")
               .build())

        program = """
Base: CV
Watch: Accumulated CV > 0.5 CV
    Mark: A
"""
        with create_engine_context(uod) as e:
            cv = e.tags["CV"]
            cv.set_value(2.0, 0)
            acc_cv = e.tags[SystemTagName.ACCUMULATED_CV]
            run_engine(e, program, 1)

            self.assertEqual(acc_cv.as_float(), 0.0)
            self.assertEqual(acc_cv.unit, "CV")

            continue_engine(e, 4)
            self.assertAlmostEqual(acc_cv.as_float(), 0.2, delta=0.1)
            self.assertEqual([], e.interpreter.get_marks())

            continue_engine(e, 4)
            self.assertAlmostEqual(acc_cv.as_float(), 0.4, delta=0.1)
            self.assertEqual([], e.interpreter.get_marks())

            continue_engine(e, 4)
            self.assertAlmostEqual(acc_cv.as_float(), 0.6, delta=0.1)
            self.assertEqual(['A'], e.interpreter.get_marks())

    def test_accumulated_column_block_volume(self):
        self.engine.cleanup()  # dispose the test default engine

        uod = (UodBuilder()
               .with_instrument("TestUod")
               .with_author("Test Author", "test@openpectus.org")
               .with_filename(__file__)
               .with_hardware(TestHW())
               .with_location("Test location")
               .with_tag(ReadingTag("CV", "L"))
               .with_tag(CalculatedLinearTag("calc", "L"))
               .with_accumulated_cv(cv_tag_name="CV", totalizer_tag_name="calc")
               .build())

        program = """
Base: s
Block: A
    0.45 End block
0.45 Mark: A
        """
        with create_engine_context(uod) as e:
            cv = e.tags["CV"]
            cv.set_value(2.0, 0)
            acc_cv = e.tags[SystemTagName.ACCUMULATED_CV]
            block_cv = e.tags[SystemTagName.BLOCK_CV]
            block = e.tags[SystemTagName.BLOCK]
            run_engine(e, program, 1)

            self.assertEqual(acc_cv.as_float(), 0.0)
            self.assertEqual(acc_cv.unit, "CV")
            self.assertEqual(block_cv.as_float(), 0.0)
            self.assertEqual(block_cv.unit, "CV")

            continue_engine(e, 2)  # Blank + Base
            self.assertEqual(block.get_value(), None)
            self.assertAlmostEqual(acc_cv.as_float(), 0.2/2, delta=0.1)
            self.assertAlmostEqual(block_cv.as_float(), 0.2/2, delta=0.1)

            continue_engine(e, 1)  # Block
            self.assertEqual(block.get_value(), "A")
            self.assertAlmostEqual(acc_cv.as_float(), 0.3/2, delta=0.1)
            self.assertAlmostEqual(block_cv.as_float(), 0.1/2, delta=0.1)

            continue_engine(e, 5)
            self.assertEqual(block.get_value(), "A")
            self.assertAlmostEqual(acc_cv.as_float(), 0.8/2, delta=0.1)
            self.assertAlmostEqual(block_cv.as_float(), 0.6/2, delta=0.1)

            continue_engine(e, 1)
            self.assertEqual(block.get_value(), None)
            # acc_vol keeps counting
            self.assertAlmostEqual(acc_cv.as_float(), 0.9/2, delta=0.1)
            # block_vol is reset to value before block A - so it matches acc_vol again
            self.assertAlmostEqual(block_cv.as_float(), 0.9/2, delta=0.1)


    # --- Units ---


    def test_units(self):
        self.engine.cleanup()  # dispose the test default engine

        uod = (UodBuilder()
               .with_instrument("TestUod")
               .with_author("Test Author", "test@openpectus.org")
               .with_filename(__file__)
               .with_hardware(TestHW())
               .with_location("Test location")
               .with_tag(ReadingTag("X1", "L"))
               .with_tag(ReadingTag("X2", "g"))
               .with_tag(ReadingTag("X3", "%"))
               .build())

        with self.subTest("test units"):
            with create_engine_context(uod) as e:
                run_engine(e, "", 5)
                self.assertEqual(e.tags[SystemTagName.METHOD_STATUS].get_value(), MethodStatusEnum.OK)

                percent_tag = uod.tags["X3"]
                self.assertEqual(percent_tag.unit, "%")


    # --- Restart ---


    def test_restart_can_restart(self):
        set_engine_debug_logging()
        set_interpreter_debug_logging()
        program = """
Mark: A
Increment run counter
Restart
"""
        e = self.engine
        self.assertEqual(0, e.tags[SystemTagName.RUN_COUNTER].as_number())

        run_engine(e, program, 5)

        run_id_1 = e.tags[SystemTagName.RUN_ID].get_value()
        self.assertEqual(e.tags[SystemTagName.SYSTEM_STATE].get_value(), SystemStateEnum.Restarting)
        self.assertEqual(1, e.tags[SystemTagName.RUN_COUNTER].as_number())

        self.assertEqual(e.interpreter.get_marks(), ["A"])

        continue_engine(e, 1)

        self.assertEqual(e.tags[SystemTagName.SYSTEM_STATE].get_value(), SystemStateEnum.Stopped)
        self.assertIsNone(e.tags[SystemTagName.RUN_ID].get_value())
        self.assertEqual(e.interpreter.get_marks(), [])
        self.assertEqual(1, e.tags[SystemTagName.RUN_COUNTER].as_number())

        continue_engine(e, 1)

        run_id2 = e.tags[SystemTagName.RUN_ID].get_value()
        self.assertEqual(e.tags[SystemTagName.SYSTEM_STATE].get_value(), SystemStateEnum.Running)
        self.assertNotEqual(run_id_1, run_id2)
        self.assertEqual(e.interpreter.get_marks(), [])
        self.assertEqual(1, e.tags[SystemTagName.RUN_COUNTER].as_number())

        continue_engine(e, 3)

        self.assertEqual(e.interpreter.get_marks(), ["A"])
        self.assertEqual(2, e.tags[SystemTagName.RUN_COUNTER].as_number())

    def test_restart_stop_ticking_interpreter(self):

        set_interpreter_debug_logging()
        program = """
Mark: A
Restart
Mark: X
"""
        e = self.engine
        run_engine(e, program, 1)

        for _ in range(30):
            continue_engine(e, 1)
            self.assertTrue("X" not in e.interpreter.get_marks())

    def test_restart_cancels_running_commands(self):
        program = """
Reset
Restart
"""
        e = self.engine
        run_engine(e, program, 4)

        # TODO look into logs - something is not right

        system_state = e.tags[SystemTagName.SYSTEM_STATE]
        self.assertEqual(system_state.get_value(), SystemStateEnum.Restarting)

    def test_EngineCommandEnum_has(self):
        self.assertTrue(EngineCommandEnum.has_value("Stop"))
        self.assertFalse(EngineCommandEnum.has_value("stop"))
        self.assertFalse(EngineCommandEnum.has_value("STOP"))

# ----------- Engine Error -------------

    def test_engine_error_causes_Paused_state(self):
        e = self.engine
        with self.assertRaises(ValueError):
            run_engine(e, "foo bar", 3)
        self.assertTrue(e.has_error_state())

    def test_interpreter_error_causes_Paused_state(self):
        e = self.engine
        with self.assertRaises(ValueError):
            run_engine(e, """WATCH x > 2""", 3)
        self.assertTrue(e._runstate_paused)


# Test helpers


class TestHW(HardwareLayerBase):
    __test__ = False

    def __init__(self) -> None:
        super().__init__()
        self.register_values = {}

    @override
    def read(self, r: Register) -> Any:
        if r.name in self.registers.keys() and r.name in self.register_values.keys():
            return self.register_values[r.name]
        return None

    @override
    def write(self, value: Any, r: Register):
        if r.name in self.registers.keys():
            self.register_values[r.name] = value

    @override
    def connect(self):
        self._is_connected = True

    @override
    def disconnect(self):
        self._is_connected = False


class CalculatedLinearTag(Tag):
    """ Test tag that is used to simulate a value that is a linear function of time. """
    def __init__(self, name: str, unit: str | None, slope: float = 1.0) -> None:
        super().__init__(name, value=0.0, unit=unit, direction=TagDirection.NA)
        self.slope = slope

    def on_start(self, run_id: str):
        self.value = time.time() * self.slope

    def on_tick(self, tick_time: float, increment_time: float):
        self.value = time.time() * self.slope


if __name__ == "__main__":
    unittest.main()



""" Event subsystem that defines life-time events during method execution. """
from __future__ import annotations

from dataclasses import dataclass
import logging
from typing import Iterable

logger = logging.getLogger(__name__)


class EventListener:
    """ Defines the listener interface and base class for engine life-time events. """
    def __init__(self) -> None:
        self.run_id: str | None = None

    def on_engine_configured(self):
        """ Invoked once on engine startup, after configuration and after
        the connection to hardware has been established. """
        pass

    def _on_before_start(self, run_id: str):
        """ Is run by emitter to avoid having all subclasses call super().on_start(). """
        # An alternative approach would be to force subclasses to call super().on_start() using a meta class.
        # This is not supported by pyright though, so it would be a custom check at runtime, e.g.
        # https://stackoverflow.com/questions/67661091/how-can-i-make-it-a-requirement-to-call-super-from-within-an-abstract-method-o
        # or
        # https://stackoverflow.com/questions/66581157/how-to-enforce-mandatory-parent-method-call-when-calling-child-method
        self.run_id = run_id

    def on_start(self, run_id: str):
        """ Is invoked by the Start command when method is started. """
        pass

    def on_block_start(self, block_info: BlockInfo):
        """ Invoked just after a new block is started, before on_tick,
            in the same engine tick.
        """
        pass

    def on_block_end(self, block_info: BlockInfo, new_block_info: BlockInfo | None):
        """ Invoked just after a block is completed, before on_tick,
        in the same engine tick. """
        pass

    def on_tick(self, tick_time: float, increment_time: float):
        """ Is invoked on each tick.

        Intended for NA (calculated/derived) tags to calculate the
        value for the tick and apply it to the value property. """
        pass

    def on_method_end(self):
        """ Is invoked when method interpretation is complete. """
        pass

    def on_stop(self):
        """ Is invoked by the Stop command when method is stopped. """
        self.run_id = None

    def on_engine_shutdown(self):
        """ Invoked once when engine shuts down"""
        self.run_id = None


class EventEmitter:
    """ Emitter of EventListener events.

    Enables emitting execution events to a collection of listeners.
    """
    def __init__(self, listeners: Iterable[EventListener]) -> None:
        self._listeners: list[EventListener] = list(listeners)

    def add_listener(self, listener: EventListener):
        self._listeners.append(listener)

    def emit_on_engine_configured(self):
        for element in self._listeners:
            try:
                element.on_engine_configured()
            except Exception:
                logger.error(f"on_engine_configured failed for element '{str(element)}'", exc_info=True)

    def emit_on_start(self, run_id: str):
        for element in self._listeners:
            try:
                element._on_before_start(run_id)
            except Exception:
                logger.error(f"on_before_start failed for element '{str(element)}'", exc_info=True)
        for element in self._listeners:
            try:
                element.on_start(run_id)
            except Exception:
                logger.error(f"on_engine_start failed for element '{str(element)}'", exc_info=True)

    def emit_on_tick(self, tick_time: float, increment_time: float):
        for element in self._listeners:
            try:
                element.on_tick(tick_time, increment_time)
            except Exception:
                logger.error(f"on_tick failed for element '{str(element)}'", exc_info=True)

    def emit_on_block_start(self, block_name: str, tick: int):
        for element in self._listeners:
            try:
                element.on_block_start(BlockInfo(block_name, tick))
            except Exception:
                logger.error(f"on_block_start failed for element '{str(element)}'", exc_info=True)

    def emit_on_block_end(self, block_name: str, new_block_name: str, tick: int):
        for element in self._listeners:
            try:
                element.on_block_end(BlockInfo(block_name, tick), BlockInfo(new_block_name, tick))
            except Exception:
                logger.error(f"on_block_end failed for element '{str(element)}'", exc_info=True)

    def emit_on_method_end(self):
        for element in self._listeners:
            try:
                element.on_method_end()
            except Exception:
                logger.error(f"on_method_end failed for element '{str(element)}'", exc_info=True)

    def emit_on_stop(self):
        for element in self._listeners:
            try:
                element.on_stop()
            except Exception:
                logger.error(f"on_stop failed for element '{str(element)}'", exc_info=True)

    def emit_on_engine_shutdown(self):
        for element in self._listeners:
            try:
                element.on_engine_shutdown()
            except Exception:
                logger.error(f"on_engine_shutdown failed for element '{str(element)}'", exc_info=True)


@dataclass(frozen=True)
class BlockInfo:
    name: str
    tick: int

    @property
    def key(self) -> str :
        """ Gets the block name as a unique name across a single excution. """
        return self.name + "_" + str(self.tick)



import json
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Any

from pydantic_core import to_jsonable_python
from sqlalchemy import create_engine, Engine
from sqlalchemy.orm import Session, sessionmaker
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.types import ASGIApp

_engine: Engine | None = None
_sessionmaker: sessionmaker[Session] | None = None
_session_ctx: ContextVar[Session | None] = ContextVar("_session", default=None)


@contextmanager
def create_scope():
    """ Used from middleware to create a request context (scope), in which scoped_session() will work.

    Can also be used in non-request contexts to enable scoped_session() there.

    Also used in tests to provide a scope
    """
    if _sessionmaker is None:
        raise DatabaseNotConfiguredError()
    session = _sessionmaker()
    token = _session_ctx.set(session)
    try:
        yield
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()
        _session_ctx.reset(token)


def scoped_session() -> Session:
    if _engine is None:
        raise DatabaseNotConfiguredError()
    session = _session_ctx.get()
    if session is None:
        raise SessionMissingError()
    return session


def configure_db(database_url: str):
    global _engine, _sessionmaker
    _engine = create_engine(
        database_url,
        echo=False,
        connect_args={"check_same_thread": False},
        json_serializer=json_serialize,
        json_deserializer=json_deserialize)
    _sessionmaker = sessionmaker(autocommit=False, autoflush=False, bind=_engine)



def json_serialize(instance) -> str:
    return json.dumps(instance, default=to_jsonable_python)


def json_deserialize(instance) -> dict[str, Any]:
    return json.loads(instance)


class DBSessionMiddleware(BaseHTTPMiddleware):
    """ FastAPI middleware that makes request scoped database sessions available """
    def __init__(self, app: ASGIApp):
        super().__init__(app)

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint):
        if _sessionmaker is None:
            raise DatabaseNotConfiguredError()
        else:
            with create_scope():
                response = await call_next(request)
            return response


class DatabaseNotConfiguredError(Exception):
    """Excetion raised when trying to access the database before it is configured. """
    def __init__(self, *args: object) -> None:
        super().__init__("Database has not been configured, i.e. configure_db() was not called.")


class SessionMissingError(Exception):
    """Excetion raised when trying to access a database session without a scope."""
    def __init__(self):
        super().__init__("No session is available in the current scope.")




from openpectus.aggregator.routers.dto import CommandExample

examples_demo = [
    CommandExample(name="Some Example", example="Some example text"),
    CommandExample(name="Watch Example", example="""
Watch: Block Time > 3s
    Mark: A
Mark: X

Watch: Block Time > 7s
    Mark: B
Mark: Y""")
]

examples = [
    CommandExample(name="Alarm", example="""
# Alarm is a block that executes when a condition is satisfied repeatedly.

# Example 1
Alarm: Block Time > 5 s
    End block

Block: A
    Mark: A

Block: B
    Mark: B

Block: C
    Mark: C

Stop

# Example 2
Base: s
Batch: Hello

Alarm: Batch = Hello
    3.0 Batch: User

Alarm: Batch = User
    3.0 Batch: Hello
"""),
    CommandExample(name="Base", example="""
# Example 1:
Base: CV # Measure time in column volumes

# Example 2:
Base: min # Measure time in minutes

# Example 3:
Base: s # Measure time in seconds
"""),
#     CommandExample(name="Batch", example="""
# """),
    CommandExample(name="Block", example="""
# Example 1:
Block: Elution
    Mark: Start elution
    End block
    # Open proper valves etc
"""),
    CommandExample(name="End block", example="""
# End the block currently executing.

# Example 1
Block: This is a block
    1.0 End block
# Block is no more.

# Example 2
Block: Layer 1
    Block: Layer 2
        End block
    # We have escaped the block "Layer 2"
    # but are still in "Layer 1".
"""),
    CommandExample(name="End blocks", example="""
# End all currently executing blocks.

# Example 1
Block: This is a block
    1.0 End block
# Block is no more.

# Example 2
Block: Layer 1
    Block: Layer 2
        End blocks
# We have escaped both blocks.
"""),
    CommandExample(name="Error", example="""
# Make a statement to the log file.

Error: Hello.
"""),
    CommandExample(name="Hold", example="""
# Hold current output values but do not
# progress with respect to time.

# Example 1
# Open outlet VA38 for one minute, then VA37.
Outlet: VA38
Hold: 1 min
Outlet: VA37
"""),
    CommandExample(name="Increment run counter", example="""
# Add 1 to the Run Counter.

# Example 1
Watch: Run Counter > 3
    Stop

Wait: 10 s

Increment run counter

Restart
"""),
    CommandExample(name="Info", example="""
# Make a statement to the log file.

Info: Hello.
"""),
    CommandExample(name="Mark", example="""
# Make a note in the data file at the current point in time.

# Example 1
Wait: 5 s
Mark: 5 seconds have passed.
"""),
    CommandExample(name="Pause", example="""
# Pause execution of commands
# and time. Put outputs into safe state.

# Example 1
Outlet: VA38
Pause: 1 min # Outlet is closed for 1 minute.
"""),
    CommandExample(name="Restart", example="""
# Stop execution, then start the method again.

# Example 1
Restart
"""),
    CommandExample(name="Run counter", example="""
# Assign a value to the Run Counter.

# Example 1
Run counter: 3
"""),
#     CommandExample(name="Simulate", example="""
# # Simulate a value of a measured tag.

# # Example 1
# Simulate: UV01 Absorbance = 1 AU
# """),
#     CommandExample(name="Simulate off", example="""
# # Stop simulation of measured tag.

# # Example 1
# Simulate: UV01 Absorbance = 1 AU
# Wait: 5s
# Simulate off: UV01 Absorbance
# """),
    CommandExample(name="Stop", example="""
# Stop execution.

# Example 1
Stop
"""),
    CommandExample(name="Wait", example="""
# Holds execution of commands for the specified duration.

# Example 1
# Open outlet VA38 for one minute, then VA37.
Outlet: VA38
Wait: 1 min
Outlet: VA37
"""),
    CommandExample(name="Warning", example="""
# Make a statement to the log file.

Warning: Hello.
"""),
    CommandExample(name="Watch", example="""
# Watch is a block that executes when a condition is satisfied.

# Example 1:
Watch: Block Time > 1 min
    Mark: One minute has passed

# Example 2:
Watch: Run Counter > 3
    Stop # Stop after 3 runs have been executed.
Wait: 5 s
Restart
"""),
]



""" Open Pectus is a process control system for Unit Operations such as filtration, chromatography, precipitation,
solubilization and refolding. Open Pectus implements a language called P-code which is used to write methods and
control components on the Unit Operation. """
__version__ = "0.1.996"

import logging
import json
import os

import colorlog

build_number = "build_number not set"
build_info = {}

path = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(path, "build.json")) as file:
    dct = json.load(file)
    build_info = dct
    build_number = dct["build_number"]


def log_setup_colorlog(root_loglevel: int = logging.INFO):
    handler = colorlog.StreamHandler()
    handler.setFormatter(colorlog.ColoredFormatter('%(log_color)s%(levelname)s:%(name)s:%(message)s'))

    logging.root.setLevel(root_loglevel)
    logging.root.handlers.clear()
    logging.root.addHandler(handler)



"""add engine_hardware_str to RecentRun model

Revision ID: 95bfa6af85ea
Revises: ec559277c8da
Create Date: 2024-02-02 12:56:53.377591

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '95bfa6af85ea'
down_revision: Union[str, None] = 'ec559277c8da'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('RecentRuns', schema=None) as batch_op:
        batch_op.add_column(sa.Column('engine_hardware_str', sa.String(), nullable=False, server_default='N/A'))
        batch_op.alter_column('engine_hardware_str', server_default=None)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('RecentRuns', schema=None) as batch_op:
        batch_op.drop_column('engine_hardware_str')

    # ### end Alembic commands ###



"""change computer_name to engine_computer_name

Revision ID: 72449425de7c
Revises: 646d56d49b76
Create Date: 2024-01-30 10:42:08.070364

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '72449425de7c'
down_revision: Union[str, None] = '646d56d49b76'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('RecentRuns', 'computer_name', new_column_name='engine_computer_name')
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('RecentRuns', 'engine_computer_name', new_column_name='computer_name')
    # ### end Alembic commands ###



from __future__ import annotations
import importlib
import logging
from typing import Any, Callable, Generator
from openpectus.engine.commands import EngineCommand
from openpectus.engine.models import EngineCommandEnum


logger = logging.getLogger(__name__)


class InternalCommandsRegistry:
    def __init__(self, engine):
        self.engine = engine
        self._command_map: dict[str, Callable[[], InternalEngineCommand]] = {}
        self._command_instances: dict[str, InternalEngineCommand] = {}

    def __enter__(self):
        self._register_commands(self.engine)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        logger.debug("Disposing registry")
        self._command_map.clear()
        for cmd in self._command_instances.values():
            try:
                cmd.cancel()
            except Exception as ex:
                logger.warning(f"Error during cancel/finalize while disposing command: {cmd.name}: {str(ex)}")
        self._finalize_instances()

    def _register_commands(self, engine):
        if len(self._command_map) > 0:
            raise ValueError("Register may only run once per engine lifetime. " +
                             "Engine was probably not cleaned up with .cleanup()")

        def register(command_name: str, cls):
            self._command_map[command_name] = lambda : cls(engine, self)
            logger.debug(f"Registered command '{command_name}' with factory '{str(cls)}'")

        # use dynamic import for introspection to avoid static import cycle error
        command_impl_module = importlib.import_module("openpectus.engine.internal_commands_impl")

        # use introspection to auto-register classes in command_impl_module
        registered_classes = []
        for command_name in EngineCommandEnum:
            for cls_name in dir(command_impl_module):
                if cls_name == f"{command_name}EngineCommand":
                    cls = getattr(command_impl_module, cls_name, None)
                    if cls is not None:
                        # register via a function here to obtain the right closure
                        register(command_name, cls)
                        registered_classes.append(cls_name)
        logger.debug(f"Registered internal engine commands: {', '.join(registered_classes)}")

    def get_running_internal_command(self) -> InternalEngineCommand | None:
        if len(self._command_instances) > 0:
            for cmd in self._command_instances.values():
                return cmd

    def create_internal_command(self, command_name: str) -> InternalEngineCommand:
        if command_name not in self._command_map.keys():
            raise ValueError(f"Command name '{command_name}' is not a known internal engine command")
        if command_name in self._command_instances.keys():
            raise ValueError(f"There is already a command instance for command name '{command_name}'")

        instance = self._command_map[command_name]()
        self._command_instances[command_name] = instance
        logger.debug(f"Created instance {type(instance).__name__} for command '{command_name}'")
        return instance

    def dispose_command(self, command_name):
        if command_name in self._command_instances.keys():
            del self._command_instances[command_name]
        else:
            logger.warning(f"No command '{command_name}' found to dispose. " +
                           f"Actual commands: {str(list(self._command_instances.keys()))}")

    def _finalize_instances(self):
        """ Finalize and dispose all command instances. """
        instances = list(self._command_instances.values())
        for cmd in instances:
            cmd.finalize()


class InternalEngineCommand(EngineCommand):
    """ Base class for internal engine commands.

    Adds support for long-running commands via a _run() generator method. The tick() base class method
    implements the state management of these commands.
    """
    def __init__(self, name: str, registry: InternalCommandsRegistry) -> None:
        super().__init__(name)
        self._registry = registry
        self._failed = False
        self.has_run = False
        self.run_result: Generator[None, None, None] | None = None
        self.kvargs: dict[str, Any] = {}

    def _run(self) -> Generator[None, None, None] | None:
        """ Override to implement the command using a generator style
        where each yield pauses execution until the next tick (i.e. call to execute()). """
        raise NotImplementedError()

    def init_args(self, kvargs: dict[str, Any]):
        self.kvargs = kvargs

    def fail(self):
        self._failed = True

    def has_failed(self):
        return self._failed

    def set_complete(self):
        super().set_complete()

    def finalize(self):
        super().finalize()
        self._registry.dispose_command(self.name)

    def tick(self) -> None:
        if self.is_finalized() or self.is_cancelled() or self.has_failed():
            # command should already be disposed and not end up here
            logger.warning(f"Did not expect internal command '{self.name}' here in its end state")
        elif self.is_execution_complete():
            self.finalize()
        else:
            if not self.has_run:
                try:
                    self.run_result = self._run()
                    self.has_run = True
                except Exception:
                    self.fail()
                    self._registry.dispose_command(self.name)
                    logger.error(f"Command '{self.name}' failed", exc_info=True)
                    return

            if self.run_result is None:
                self.set_complete()
                self.finalize()
            else:
                try:
                    next(self.run_result)
                except StopIteration:
                    self.set_complete()
                    self.finalize()
                except Exception:
                    self.fail()
                    self.finalize()
                    logger.error(f"Command '{self.name}' failed", exc_info=True)
